import yaml
import shutil
import os
import tempfile
import numpy as np
from typing import Dict, List
from pathlib import Path
from torch import nn
from ngclib import NGCDir
from ngclib.utils import generateRandomData
from ngclib.trainer import NGCTrainer
from ngclib.nodes import *
from ngclib.models import getModel
from nwgraph import Node

tmpDirName = None

graphV1CfgStr = """
graphConfig: NGC-V1
nodeTypes: [RGB, RGB, RGB, RGB, Semantic]
nodeNames: [rgb1, rgb2, rgb3, rgb4, semantic]
inputNodes: [rgb1, rgb2, rgb3, rgb4]
edges: [ # Edges are defined using the names of the nodes
    [rgb1, semantic],
    [rgb2, semantic],
    [rgb3, semantic],
    [rgb4, semantic],
]
voteFunction: simpleMean
hyperParameters:
  semantic:
    semanticClasses: 12
    semanticColors: [
      [0, 0, 0],
      [0, 0, 1],
      [0, 0, 2],
      [0, 0, 3],
      [0, 0, 4],
      [0, 0, 5],
      [0, 0, 6],
      [0, 0, 7],
      [0, 0, 8],
      [0, 0, 9],
      [0, 0, 10],
      [0, 0, 11],
    ]
"""

trainCfgStr = """
seed: 42
cudnn_determnisim: True
batchSize: 10
numEpochs: 3
optimizer:
  type: adamw
  args:
    lr: 0.01
scheduler:
  type: ReduceLROnPlateau
  args:
    metric: Loss
    patience: 10
    factor: 2
"""

def generateData():
    # Generate data
    global tmpDirName
    if tmpDirName is None:
        if __name__ == "__main__":
            tmpDirName = Path("/tmp/tmp8mddpuw8")
        else:
            tmpDirName = Path(tempfile.TemporaryDirectory().name)
    print(tmpDirName)
    if not (tmpDirName / "iter1/data").exists():
        generateRandomData(baseDir=tmpDirName / "iter1/data", representations=["rgb1", "semantic"], \
            dims=[3, 12], types=["float", "categorical"], resolution=(240, 426), N=20, prefix="train_")
    if not (tmpDirName / "validation").exists():
        generateRandomData(baseDir=tmpDirName / "validation", representations=["rgb1", "semantic"], \
            dims=[3, 12], types=["float", "categorical"], resolution=(240, 426), N=5)
    if not (tmpDirName / "semisupervised").exists():
        generateRandomData(baseDir=tmpDirName / "semisupervised", representations=["rgb1"], dims=[3, 3], \
            types=["float"], resolution=(240, 426), N=10)

    def f(a, b):
        if not (tmpDirName / b).exists():
            os.symlink(str(tmpDirName / a), str(tmpDirName / b))
    f((tmpDirName / "iter1/data/rgb1"), (tmpDirName / "iter1/data/rgb2"))
    f((tmpDirName / "iter1/data/rgb1"), (tmpDirName / "iter1/data/rgb3"))
    f((tmpDirName / "iter1/data/rgb1"), (tmpDirName / "iter1/data/rgb4"))
    f((tmpDirName / "validation/rgb1"), (tmpDirName / "validation/rgb2"))
    f((tmpDirName / "validation/rgb1"), (tmpDirName / "validation/rgb3"))
    f((tmpDirName / "validation/rgb1"), (tmpDirName / "validation/rgb4"))
    f((tmpDirName / "semisupervised/rgb1"), (tmpDirName / "semisupervised/rgb2"))
    f((tmpDirName / "semisupervised/rgb1"), (tmpDirName / "semisupervised/rgb3"))
    f((tmpDirName / "semisupervised/rgb1"), (tmpDirName / "semisupervised/rgb4"))
    return tmpDirName

class IdentityNode(Node):
    def __init__(self, numDims: int):
        self.numDims = numDims

    def getEncoder(self, decoderNode):
        return nn.Linear(in_features = self.numDims, out_features = decoderNode.numDims)

    def getDecoder(self, encoderNode):
        return nn.Identity()

class _RGB(RGB, IdentityNode):
    def __init__(self, numDims, *args, **kwargs):
        RGB.__init__(self, *args, **kwargs)
        IdentityNode.__init__(self, numDims)

class _Semantic(Semantic, IdentityNode):
    def __init__(self, numDims, *args, **kwargs):
        Semantic.__init__(self, *args, **kwargs)
        IdentityNode.__init__(self, numDims)

class _Depth(Depth, IdentityNode):
    def __init__(self, numDims, *args, **kwargs):
        Depth.__init__(self, *args, **kwargs)
        IdentityNode.__init__(self, numDims)

def getNodes(cfg: Dict) -> List[Node]:
    """Basic implementation of getNodes, creating a subtype for all required types by adding a linear projection
    for the encoder and identity for decoder.
    """
    nodes = []
    for i in range(len(cfg["nodeNames"])):
        nodeName = cfg["nodeNames"][i]
        nodeParams = {} if not nodeName in cfg["hyperParameters"] else cfg["hyperParameters"][nodeName]
        nodeParams["name"] = nodeName
        if nodeName in ("rgb1", "rgb2", "rgb3", "rgb4"):
            node = _RGB(3, **nodeParams)
        elif nodeName == "semantic":
            node = _Semantic(nodeParams["semanticClasses"], **nodeParams)
        else:
            assert False, nodeName
        nodes.append(node)
    return nodes

def cleanup(baseDir: Path):
    shutil.rmtree(baseDir/"iter1/models") if (baseDir/"iter1/models").exists() else None
    shutil.rmtree(baseDir/"iter2") if (baseDir/"iter2").exists() else None
    shutil.rmtree(baseDir/"iter3") if (baseDir/"iter3").exists() else None

def train(trainCfg, graphCfg):
    baseDir = generateData()
    cleanup(baseDir)
    ngcDir = NGCDir(baseDir, graphCfg)
    print(ngcDir)

    nodes = getNodes(graphCfg)
    print(f"Nodes: {nodes}")

    model = getModel(nodes, graphCfg)
    print(model.summary())

    trainer = NGCTrainer(model=model, ngcDirPath=baseDir, method="export_new_dataset_every_iteration", \
        trainCfg=trainCfg, trainDir=baseDir / "iter1/data", validationDir=baseDir / "validation", \
        semisupervisedDirs=[baseDir / "semisupervised"])
    trainer.run(startIteration=1, endIteration=2, parallelization=False, debug=False)

    cleanup(baseDir)
    return model

class TestReproductibility:
    def test_reproductibility_1(self):
        trainCfg = yaml.safe_load(trainCfgStr)
        graphCfg = yaml.safe_load(graphV1CfgStr)
        model = train(trainCfg, graphCfg)

        new_data = np.random.randn(10, 240, 426, 3).astype(np.float32)
        res = [edge.npForward(new_data) for edge in model.edges]
        assert np.std(res, axis=0).sum() <= 1e-3

if __name__ == "__main__":
    TestReproductibility().test_reproductibility_1()



from typing import List, Dict

def edgeYamlStrToDirName(edgeYamlStr: List[str], graphCfg: Dict) -> str:
    """Given a yaml format string ([rgb, depth, semantic]) and a modelType (NGC_V1, NGC_Next etc.), returns the
    name of the dir inside ngcdir.
    """
    modelType = graphCfg["graphConfig"]
    if modelType == "NGC-V1":
        if len(edgeYamlStr) == 2:
            return f"SL_{edgeYamlStr[0]}_{edgeYamlStr[1]}"
        elif len(edgeYamlStr) == 3:
            return f"TH_NOGT_{edgeYamlStr[0]}_{edgeYamlStr[1]}_{edgeYamlStr[2]}"
    elif modelType == "NGC_AAAI2020":
        edge = edgeYamlStr.split("_")
        assert len(edge) == 2
        if edge[0] == graphCfg["inputNode"]:
            return f"SL_{edge[0]}_{edge[1]}"
        else:
            return f"TH_NOGT_{edge[0]}_{edge[1]}"
    elif modelType == "NGC-V2-Residual":
        if len(edgeYamlStr) == 2:
            if edgeYamlStr[0] == graphCfg["residualNode"]:
                return f"SL_{edgeYamlStr[0]}_{edgeYamlStr[1]}"
            else:
                return f"SLR_{graphCfg['residualNode']}+{edgeYamlStr[0]}_{edgeYamlStr[1]}"
        elif len(edgeYamlStr) == 3:
            return f"THR_NOGT_{graphCfg['residualNode']}+{edgeYamlStr[0]}_{edgeYamlStr[1]}_{edgeYamlStr[2]}"
    else:
        assert False, f"Unknown type {modelType}."



import torch as tr
import numpy as np
from pathlib import Path
from natsort import natsorted
from typing import List, Dict
from nwgraph import Node
from torch.utils.data import Dataset, DataLoader, Subset
from collections import OrderedDict
from media_processing_lib.image import image_resize
from .transforms import defaultTransform, semanticSegmentationTransform, depthTransform
from ..nodes import *
from ..logger import logger
from ..utils import loadNpz

def getDataTransform(node):
	if isinstance(node, RGB):
		return defaultTransform
	elif isinstance(node, Depth):
		return depthTransform
	elif isinstance(node, Semantic):
		return semanticSegmentationTransform
	elif isinstance(node, CameraNormal):
		return defaultTransform
	elif isinstance(node, Normal):
		return defaultTransform
	assert False, f"Unknown node: '{node}'"

class NGCNpzReader(Dataset):
	def __init__(self, path: Path, nodes: List[Node], gtNodes: List[Node]):
		self.path = Path(path).absolute()
		self.inFiles = self.buildDataset(self.path, nodes)
		self.transforms = {x.name: getDataTransform(x) for x in nodes}
		self.nodes = nodes
		self.gtNodes = gtNodes
		self.nameToNode = {x.name: x for x in nodes}

	# For Npy reader, we expect a structure of:
	# baseDir/
	#   dataDim1/0.npz, ..., N.npz
	#   ...
	#   dataDimM/0.npz, ..., N.npz
	def buildDataset(self, path: Path, nodes: List[Node]) -> Dict[str, List[Path]]:
		logger.debug(f"Building dataset from: '{path}'")
		logger.debug(f"Nodes: {nodes}")
		inFiles = OrderedDict()
		for node in nodes:
			Dir = path / node.name
			items = [x for x in Dir.glob("*.npz")]
			items = natsorted([str(x) for x in items])
			inFiles[node.name] = items
		Lens = [len(x) for x in inFiles.values()]
		assert np.std(Lens) == 0, f"Lens: {dict(zip(nodes, Lens))}"
		assert len(inFiles) > 0
		logger.debug(f"Found {Lens[0]} images")
		return inFiles

	def getNodeItem(self, node: Node, index: int) -> np.ndarray:
		item = loadNpz(self.inFiles[node.name][index])
		item = image_resize(item, height=240, width=426, only_uint8=False)
		transformedItem = self.transforms[node.name](item, node)
		return transformedItem

	def __getitem__(self, index: int):
		result = {}
		# By default we put all nodes' data in "data" key
		for node in self.nodes:
			result[node.name] = self.getNodeItem(node, index)
		return result

	def __len__(self) -> int:
		return len(self.inFiles[list(self.nodes)[0].name])

	def mergeFn(self, x):
		# Basic merge function. Just stack all arrays received from __getitem__(ix) and return same thing in data and
		#  labels dicts for further use.
		Keys = x[0].keys()
		res = {k: np.array([y[k] for y in x], dtype=np.float32) for k in Keys}
		return {"data": res, "labels": res}

	def toDataLoader(self, batchSize: int=1, randomize: bool=False, numWorkers: int=None, \
			debug: bool=False, seed: int=None):
		reader = self
		numWorkers = 4 if numWorkers is None and debug is False else 0
		g = None
		if debug:
			reader = Subset(reader, list(range(0, 5)))
		if seed is not None:
			g = tr.Generator()
			g.manual_seed(seed)
		loader = DataLoader(reader, batch_size=batchSize, shuffle=randomize, \
			collate_fn=self.mergeFn, num_workers=numWorkers, generator=g)
		return loader



import torch as tr
from abc import abstractmethod
from pathlib import Path
from typing import List, Optional, Dict, Iterable, Callable
from nwgraph import Graph, Edge, Node, Message
from overrides import overrides
from ..nodes import *
from ..voting_algorithms import simpleMean, simpleMedian, NetworkSelection
from ..logger import logger

class NGC(Graph):
    def __init__(self, nodes: List[Node], cfg: Dict):
        self.cfg = cfg
        self.nodes = nodes
        edges = self.getNGCEdges()
        # Edge and node hyperparameters are set by graph constructor
        hyperParameters = {
            "nodeTypes": cfg["nodeTypes"],
            "voteFunction": cfg["voteFunction"]
        }
        super().__init__(edges, hyperParameters)
        assert len(edges) == len(self.edges)
        self.voteFn = self.setupVoteFn()

    def setupVoteFn(self):
        """Setup the voting function for each node, read from the cfg file """
        def getVoteFn(vote):
            if vote == "simpleMean":
                return simpleMean
            elif vote == "simpleMedian":
                return simpleMedian
            elif isinstance(vote, list) and vote[0] == "networkSelection":
                return NetworkSelection(vote)
            else:
                assert False, f"Unknown vote type: {vote}"

        vote = self.cfg["voteFunction"]
        nodeNames = [x.name for x in self.nodes]
        # We can give just one string for all nodes, or we can instantiate all of them
        if not isinstance(vote, dict):
            vote = {k: vote for k in nodeNames}

        res = {}
        for node, nodeName in zip(self.nodes, nodeNames):
            assert nodeName in vote, f"Node {nodeName} ({node}) not in vote list: {list(vote.keys())}"
            res[node] = getVoteFn(vote[nodeName])
        return res

    @overrides
    def aggregate(self):
        for node in self.nodes:
            messages = node.getMessages()
            assert len(messages) > 0
            # Concatenate all messages into one tensor
            trMessages = tr.stack([x.output for x in messages], dim=0)
            oldPath = "|".join([str(x.path[-1]) for x in messages])
            newPath = [oldPath, "Vote"]
            voteOutput = self.voteFn[node](node, trMessages)
            newMessage = Message(newPath, trMessages, voteOutput)
            node.clearMessages()
            node.addMessage(newMessage)

    @abstractmethod
    def getNGCEdges(self) -> List[Edge]:
        pass

    def getTrainedEdges(self, weightsDir: Path, getEdgeDirNameFn: Callable[[Edge], str]) -> List[Edge]:
        def isTrained(edgeWeightDir: Path) -> bool:
            if not edgeWeightDir.exists():
                return False
            # It's impossible to know if a model was fully trained or not. Thus, if this file exists, we assume that
            #  the edge is trained properly. If we want to update the edge somehow, we must retraine it separately
            #  and plug it into ngcdir.
            weightsFile = edgeWeightDir / "model_best_Loss.pkl"
            return weightsFile.exists()

        edges = []
        logger.info("Checking which edges are trained.")
        for edge in self.edges:
            if isTrained(weightsDir / getEdgeDirNameFn(edge)):
                edges.append(edge)
        logger.info(f"Trained edges: {len(edges)} out of {len(self.edges)}")
        return edges

    def loadAllEdges(self, weightsDir: Optional[Path], getEdgeDirNameFn: Callable[[Edge], str]):
        if weightsDir is None:
            logger.debug("Weights dir not set, assuming cwd()")
            weightsDir = Path.cwd()
        num_trained_edges = len(self.getTrainedEdges(weightsDir, getEdgeDirNameFn))
        assert num_trained_edges == len(self.edges), f"{num_trained_edges} vs. {len(self.edges)}"

        for edge in self.edges:
            weightsFile = weightsDir / getEdgeDirNameFn(edge) / "model_best_Loss.pkl"
            logger.debug(f"Edge: {edge}. Loading weights from '{weightsFile}'.")
            edge.setTrainableWeights(True)
            edge.loadWeights(weightsFile, yolo=True)

    def getNumParams(self):
        return "Do you really wanna know ? (laggy)", 0
        # return getNumParams(self)

    def trainReader(self, reader: Iterable, numEpochs: int, validationReader: Optional[Iterable] = None):
        from ..trainer import NGCTrainer
        assert False, "TODO: Train all automagically"

    def __str__(self):
        Str = super().__str__()
        Str += f"\nNGC:"
        Str += f"\n  - Graph type: {str(type(self)).split('.')[-1][:-2]}"
        Str += f"\n  - Vote function: {self.cfg['voteFunction']}"
        return Str



from copy import copy
from typing import List, Dict, List
from nwgraph import Node
from ..ngc_v1 import NGCV1

class NGCAAAI2020(NGCV1):
    """
    This is the original implementation. The CFG file contains the edges given manually in the format: 'nodeA_nodeB'
    It also has only one 'inputNode' entry, instead of a list on 'inputNodes' as in the NGCV1 code.
    Thus, we must convert the original cfg file to this new format.
    """
    def __init__(self, nodes:List[Node], cfg: Dict):
        assert not "inputNodes" in cfg
        cfg = NGCAAAI2020._convertCfgToV1Format(cfg)
        super().__init__(nodes, cfg)
        self.hyperParameters["inputNode"] = cfg["inputNode"]

    @staticmethod
    def _convertCfgToV1Format(cfg: Dict):
        cfg = copy(cfg)
        cfg["inputNodes"] = [cfg["inputNode"]]
        edges = cfg["edges"]
        newEdges = []
        for edge in edges:
            nodeA, nodeB = edge.split("_")
            if nodeA == cfg["inputNode"]:
                # Single Link
                newEdges.append([nodeA, nodeB])
            else:
                # Two Hops No GT
                newEdges.append([cfg["inputNode"], nodeA, nodeB])
        cfg["edges"] = newEdges
        return cfg



from nwgraph import Node

# @brief Generic Vector Node (1D) having a number of channels (DxNC)
class VectorNode(Node):
	def __init__(self, name: str, numDims: int, hyperParameters: dict={}):
		super().__init__(name, hyperParameters)
		self.numDims = numDims



from typing import Callable

from .ngc_augmentation import NGCAugmentation, GeneralPrototype, NodeSpecificPrototype
from .random_zoom import RandomZoom

def get_augmentation(name: str, **kwargs) -> Callable:
    if name == "random_zoom":
        return RandomZoom(**kwargs)

    assert False, f"Unknown augmentation: '{name}'"



from abc import abstractmethod
from typing import List
from nwgraph import SimpleEdge

class NGCEdge(SimpleEdge):
    def __init__(self, prefix: str, *args, **kwargs):
        self.prefix = prefix
        super().__init__(*args, **kwargs)

    @abstractmethod
    def getInKeys(self) -> List[str]:
        pass

    def setup(self, **kwargs):
        pass



from nwgraph import Node

class MapNode(Node):
    """Generic Map Node (2D) having a number of channels (D1xD2xNC)"""
    def __init__(self, name: str, numDims: int, hyperParameters: dict={}):
        Node.__init__(self, name, hyperParameters)
        self.numDims = numDims



import torch as tr
import numpy as np
from pathlib import Path
from typing import List, Dict, Any
from functools import lru_cache
from datetime import datetime, timedelta
from .utils import edgeYamlStrToDirName
from .str_formatter import strFormatter
from .html_formatter import htmlFormatter

# Generic ngc dir analysis class. It should be able to tell us what edges are trained, for how many iterations and
#  other analysis stuff like this.
class NGCDir:
    def __init__(self, path: Path, graphCfg: Dict):
        self.path = Path(path)
        self.graphCfg = graphCfg
        assert self.path.exists(), self.path
        assert self.graphCfg is not None

    def getNodes(self) -> Dict[str, List[str]]:
        return {"types": self.graphCfg["nodeTypes"], "names": self.graphCfg["nodeNames"]}

    def getAvailableIterations(self) -> int:
        allDirs = list(filter(lambda x: x.is_dir() and x.name.startswith("iter"), self.path.iterdir()))
        return len(allDirs)

    # @brief Returns the number of data files for each node given an iteration (train & semisupervised)
    def getDataStatus(self, iteration: int) -> Dict[str, Dict[str, int]]:
        dataDir = self.path / f"iter{iteration}/data"
        iterDataDirs = filter(lambda x: x.is_dir(), dataDir.iterdir())
        res = {}
        for iterDataDir in iterDataDirs:
            npzFiles = [x.name for x in  iterDataDir.glob("*.npz")]
            trainNpzFiles = list(filter(lambda x: x.startswith("train_"), npzFiles))
            # ss1_XX.npz, ss2_XX.npz etc.
            ssFiles = list(filter(lambda x: x.startswith("ss"), npzFiles))
            resDataDir = {
                "all": len(npzFiles),
                "train": len(trainNpzFiles),
                "semisupervised": len(ssFiles)
            }
            uniqueSSPrefixes = set([x.split("_")[0] for x in ssFiles])
            uniqueSSFiles = {}
            for uniquePrefix in uniqueSSPrefixes:
                uniqueSSFiles[uniquePrefix] = list(filter(lambda x : x.startswith(f"{uniquePrefix}_"), ssFiles))
                resDataDir[uniquePrefix] = len(uniqueSSFiles[uniquePrefix])
            res[iterDataDir.name] = resDataDir
        return res

    def checkIfTrained(self, iteration: int, edgeName: str) -> bool:
        iterDir = self.path / f"iter{iteration}/models"
        edgeDir = iterDir / edgeName
        if not edgeDir.exists():
            return False
        modelPath = edgeDir / "model_best_Loss.pkl"
        if not modelPath.exists():
            return False
        return True

    # @brief Returns a dictionary of all edges with keys as the names of the edges and values with paths to the
    #  best and last epoch (if trained) or nones.
    def getEdgesStatus(self, iteration: int) -> Dict[str, Dict[str, Path]]:
        allEdges = self.graphCfg["edges"]
        if self.graphCfg["graphConfig"] in ("NGC-V1", "NGC_AAAI2020") and allEdges == "complete":
            from ..models.ngc_v1.ngc_v1 import NGCV1
            allEdges = NGCV1._getNGCEdgesComplete(self.graphCfg["nodeNames"], self.graphCfg["inputNodes"])

        res = {}
        for edge in allEdges:
            edgeName = edgeYamlStrToDirName(edge, self.graphCfg)
            trained = self.checkIfTrained(iteration, edgeName)
            res[edgeName] = trained
        return res

    @lru_cache
    def getTrainedEdgeMetadata(self, iteration: int, edgeName: str) -> Dict[str, Any]:
        assert self.checkIfTrained(iteration, edgeName)
        iterDir = f"{self.path}/iter{iteration}/models"
        modelPath = f"{iterDir}/{edgeName}/model_last.pkl"
        historyDict = tr.load(modelPath, map_location="cpu")["history_dict"]
        res = {
            "modelPath": modelPath,
            "historyDict": historyDict
        }
        return res

    def getAllDataDirs(self, numIterations: int) -> Dict[str, Dict[int, Path]]:
        dataDirs = {i: Path(f"{self.path}/iter{i}/data") for i in range(1, numIterations + 1)}
        modelsDirs = {i: Path(f"{self.path}/iter{i}/models") for i in range(1, numIterations + 1)}
        res = {"data": dataDirs, "models": modelsDirs}
        return res

    # @brief Generic status creating method, that can be used by other services to print the status of a ngcdir.
    # @return The dicitonary that contains all the information regarding the ngcdir.
    def getStatus(self) -> Dict:
        def getDuration(historyDict):
            epoch = datetime(1900, 1, 1)
            N = len(historyDict)
            res = {}
            total = timedelta(0)
            for i in range(N):
                res[i] = {}
                item = historyDict[i]
                if not "Train" in item or not "duration" in item["Train"]:
                    res[i]["train"] = timedelta(0)
                else:
                    res[i]["train"] = datetime.strptime(item["Train"]["duration"], "%H:%M:%S.%f") - epoch

                if not "Validation" in item or not "duration" in item["Validation"]:
                    res[i]["validation"] = timedelta(0)
                else:
                    res[i]["validation"] = datetime.strptime(item["Validation"]["duration"], "%H:%M:%S.%f") - epoch
                total += res[i]["train"] + res[i]["validation"]
            res["total"] = total
            return res

        res = {
            "Path": self.path,
            "Cfg": self.graphCfg,
            "numIterations": self.getAvailableIterations(),
            "iterationInfo": {}
        }
        numIterations = self.getAvailableIterations()
        for iteration in range(1, numIterations + 1):
            iterRes = {}

            # Data stuff
            dataStatus = self.getDataStatus(iteration=iteration)
            iterRes["data"] = {}
            if len(dataStatus) == 0:
                iterRes["data"] = "n/a"
            for node, values in dataStatus.items():
                typesSemisupervised = sorted(list(filter(lambda x: x.startswith("ss"), values.keys())))
                iterRes["data"][node] = {
                    "numAll": values["all"],
                    "numTrain": values["train"],
                    "numSemisupervised": values["semisupervised"],
                    "typesSemisupervised": typesSemisupervised
                }

            # Model stuff
            iterRes["model"] = {}
            modelStatus = self.getEdgesStatus(iteration = iteration)
            iterRes["model"]["totalEdges"] = len(modelStatus)
            iterRes["model"]["totalTrained"] = sum(list(modelStatus.values()))
            iterRes["model"]["edges"] = {}
            iterRes["model"]["totalDuration"] = timedelta(0)
            for edge in modelStatus.keys():
                if modelStatus[edge] == 0:
                    iterRes["model"][edge] = {"trainedEpochs": 0, "bestEpoch": 0, "historyDict": {}, "duration": 0}
                    continue

                edgeStatus = {}
                meta = self.getTrainedEdgeMetadata(iteration, edge)
                edgeStatus["trainedEpochs"] = len(meta["historyDict"])
                losses = [meta["historyDict"][i]["Validation"][('Loss', )] for i in range(len(meta["historyDict"]))]
                edgeStatus["bestEpoch"] = np.argmin(losses) + 1
                edgeStatus["duration"] = getDuration(meta["historyDict"])

                edgeStatus["historyDict"] = meta["historyDict"]
                iterRes["model"]["edges"][edge] = edgeStatus
                iterRes["model"]["totalDuration"] += edgeStatus["duration"]["total"]
            res["iterationInfo"][iteration] = iterRes
        return res

    def __str__(self) -> str:
        return strFormatter(self.getStatus())

    def toHtml(self, *args, **kwargs) -> str:
        return htmlFormatter(self.getStatus(), *args, **kwargs)

    def __repr__(self) -> str:
        return str(self)



from .map_node import MapNode
from torch import nn

class DummyIdentityMapNode(MapNode):
    """A Dummy map node with Identity/linear layers. Mostly for testing purposes. """
    def __init__(self, name: str, numDims: int):
        super().__init__(name, numDims)

    def getEncoder(self, outputNode):
        return nn.Linear(in_features=self.numDims, out_features=outputNode.numDims)
    
    def getDecoder(self, inputNode):
        return nn.Identity()

    def aggregate(self):
        pass

    def getNodeCriterion(self):
        return lambda y, t : None
    
    def getNodeMetrics(self):
        return {}



import torch as tr
from overrides import overrides
from typing import Dict, List
from copy import deepcopy
from nwgraph import Node, Edge
from nwmodule.serializer import NWModuleSerializer
from .ngc_edge import NGCEdge
from ...logger import logger

device = tr.device("cuda") if tr.cuda.is_available() else tr.device("cpu")

class TwoHopsSerializer(NWModuleSerializer):
    @overrides
    def doSaveWeights(self):
        # Guarantee that the single link is not trainable at save time, so it's not saved.
        isTrainable = self.model.singleLinkEdge.isTrainable()
        self.model.singleLinkEdge.setTrainableWeights(False)
        parametersState = super().doSaveWeights()
        self.model.singleLinkEdge.setTrainableWeights(isTrainable)
        return parametersState

    @overrides
    def doLoadWeights(self, loadedParams, allowNamedMismatch=False):
        # Temporarily remove the reference to the single link edge from the two hops model so we can load w/o these
        #  names missing from the state_dict.
        slEdge = self.model.singleLinkEdge
        self.model.singleLinkEdge = None
        super().doLoadWeights(loadedParams, allowNamedMismatch)
        self.model.singleLinkEdge = slEdge

class TwoHopsNoGT(NGCEdge):
    def __init__(self, singleLinkEdge: Edge, inputNode: Node, outputNode: Node, hyperParameters: Dict={}):
        singleLinkNode = singleLinkEdge.inputNode
        name = f"Two Hops NoGT ({inputNode} -> {outputNode}) (Single Link Node: {singleLinkNode})"
        prefix = "TH_NOGT"
        hyperParameters["singleLinkNode"] = singleLinkNode
        super().__init__(prefix, inputNode, outputNode, name, hyperParameters)
        self.singleLinkNode = singleLinkNode
        self.singleLinkEdge = deepcopy(singleLinkEdge).to(self.getDevice())
        self.serializer = TwoHopsSerializer(model=self)
        logger.debug(f"Setting single link of {self.name}: '{self.singleLinkEdge}'")

    @overrides
    def setup(self, **kwargs):
        weightsFile = f"{kwargs['singleLinkDir']}/model_best_Loss.pkl" 
        logger.debug(f"Loading single link of {self.name}: {self.singleLinkEdge} from '{weightsFile}'")
        self.singleLinkEdge.setTrainableWeights(True)
        self.singleLinkEdge.loadWeights(weightsFile)
        self.singleLinkEdge.eval()

    @overrides
    def networkAlgorithm(self, trInputs, trLabels, isTraining, isOptimizing):
        assert not self.singleLinkEdge is None
        assert not self.getCriterion() is None, "Set criterion before training or testing"
        # Pass through the single link of this two hops edge
        # slInput = trInputs[self.singleLinkNode.name]
        with tr.no_grad():
            slOutput, _ = self.singleLinkEdge.networkAlgorithm(trInputs, trLabels, False, False)
        trResults = self.forward(slOutput)
        trLoss = 0
        if isTraining:
            trLoss = self.getCriterion()(trResults, trLabels)
            self.updateOptimizer(trLoss, isTraining, isOptimizing)
        return trResults, trLoss

    @overrides
    def getInKeys(self) -> List[str]:
        return self.singleLinkEdge.getInKeys()



from typing import Dict, List, Tuple
import io
import base64
import numpy as np
import matplotlib.pyplot as plt
from media_processing_lib.image import image_read

def npArayToBase64(arr: np.ndarray) -> str:
    plt.figure()
    plt.imshow(arr)
    res = pltToBase64(dpi=300, axis=False)
    plt.close()
    return res

def pltToBase64(dpi=None, axis: bool=True) -> str:
    f = io.BytesIO()
    if axis is False:
        plt.gca().set_axis_off()
        plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        plt.margins(0, 0)
        plt.gca().xaxis.set_major_locator(plt.NullLocator())
        plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.savefig(f, format="jpg", bbox_inches="tight", pad_inches=0, dpi=dpi)
    f.seek(0)
    jpgBase64 = str(base64.b64encode(f.read()), "utf8")
    return jpgBase64

def metricToBase64(trainData: List, valData: List, metricName: str) -> str:
    plt.figure()
    plt.plot(range(1, len(trainData) + 1), trainData, label="Train")
    plt.plot(range(1, len(valData) + 1), valData, label="Validation")
    minX, minValue = np.argmin(valData), np.min(valData)
    plt.annotate(f"Epoch {minX + 1}\nMin {minValue:2.2f}", xy=(minX + 1, minValue))
    plt.plot([minX + 1], [minValue], "o")
    plt.title(metricName)
    plt.legend()
    res = pltToBase64()
    plt.close()
    return res

def getMetricData(edgeData: Dict, metricName: str) -> Tuple[List, List]:
    trainLosses = [edgeData["historyDict"][i]["Train"][metricName] for i in range(len(edgeData["historyDict"]))]
    valLosses = [edgeData["historyDict"][i]["Validation"][metricName] for i in range(len(edgeData["historyDict"]))]
    return trainLosses, valLosses

def getBestLoss(edgeData: Dict, metricName: str) -> float:
    valLosses = [edgeData["historyDict"][i]["Validation"][metricName] for i in range(len(edgeData["historyDict"]))]
    return np.min(valLosses)

def cacheFn(path) -> str:
    """Create the stack only if needed, otherwise cache it for later usage."""
    cacheImgPath = path / "img.base64"
    if cacheImgPath.exists():
        res = open(cacheImgPath, "r").read()
    else:
        res = np.concatenate([image_read(str(x)) for x in path.glob("*.png")], axis=0)
        res = npArayToBase64(res)
        open(cacheImgPath, "w").write(res)
    return res

def button(textStr: str, imgStr: str) -> str:
    return f"""
<button class="buttonWithDiv">{textStr}</button>
<div class="hiddenDiv">
    <img src="data:image/jpg;base64, {imgStr}" />
</div>"""


def getHtmlForModels(status: Dict, iteration: int, getLoss: bool, getBestSamples: bool, getLastSamples: bool) -> str:
    edges = status["iterationInfo"][iteration]["model"]["edges"]
    if len(edges) == 0:
        return ""

    Str = """
<h3> Edges </h3>
<table>
    <thead>
        <tr>
            <th rowspan="2"> Edge </th>
            <th rowspan="1" colspan="2"> Loss </th>
            <th rowspan="1" colspan="2"> Epochs </th>
            <th rowspan="2" colspan="1"> Duration </th>
            <th rowspan="1" colspan="4"> Samples </th>
        </tr>
        <tr>
            <th> Value </th>
            <th> Toggle </th>
            <th> Total </th>
            <th> Best </th>
            <th> Best train </th>
            <th> Best validation </th>
            <th> Last train </th>
            <th> Last validation </th>
        </tr>
    </thead>
"""

    for edgeName, edgeData in edges.items():
        trainLosses, valLosses = getMetricData(edgeData, metricName=("Loss", ))

        base64Loss = button("Toggle", metricToBase64(trainLosses, valLosses, metricName="Loss")) if getLoss else "n/a"
        bestValLoss = getBestLoss(edgeData, metricName=("Loss", ))

        samplesDir = status["Path"] / f"iter{iteration}/models/{edgeName}/samples"
        bestEpochDir = samplesDir / f"{np.argmin(valLosses) + 1}"
        lastEpochDir = samplesDir / f"{len(valLosses)}"
        bestTrainSamples = button("Toggle", cacheFn(bestEpochDir / "train")) if getBestSamples else "n/a"
        bestValidationSamples = button("Toggle", cacheFn(bestEpochDir / "validation")) if getBestSamples else "n/a"
        lastTrainSamples = button("Toggle", cacheFn(lastEpochDir / "train")) if getLastSamples else "n/a"
        lastValidationSamples = button("Toggle", cacheFn(lastEpochDir / "validation")) if getLastSamples else "n/a"

        Str += f"""
<tr>
<td>{edgeName}</td>
<td>{bestValLoss:.3f}</td>
<td>
    <div> {base64Loss} </div>
</td>
<td> {edgeData['trainedEpochs']} </td>
<td> {edgeData['bestEpoch']} </td>
<td> {edgeData['duration']['total']} </td>

<td> <div> {bestTrainSamples} </div> </td>
<td> <div> {bestValidationSamples} </div> </td>
<td> <div> {lastTrainSamples} </div> </td>
<td> <div> {lastValidationSamples} </div> </td>

</tr>"""

    Str += "</table>"
    return Str

def getHtmlForData(status: Dict, iteration: int) -> str:
    dataStatus = status["iterationInfo"][iteration]["data"]
    if dataStatus == "n/a":
        return "<h3> Data </h3> <br/> n/a"

    Str = """
<h3> Data </h3>
<table>
<tr>
    <th> Node </th>
    <th> Total </th>
    <th> Train </th>
    <th> Semisupervised </th>
    <th> Identifiers </th>
</tr>
"""
    for node in dataStatus.keys():
        strTypes = ", ".join(dataStatus[node]["typesSemisupervised"])
        Str += f"""
<tr>
    <td> {node} </td>
    <td> {dataStatus[node]["numAll"]} </td>
    <td> {dataStatus[node]["numTrain"]} </td>
    <td> {dataStatus[node]["numSemisupervised"]} </td>
    <td> {strTypes} </td>
</tr>
"""
    Str += "</table>"
    return Str

def js() -> str:
    Str = "<script src=\"https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js\"></script>"
    Str += """
<script>

function toggleHideShow(item) {
    if (item.is(":hidden")) {
        item.show();
    }
    else {
        item.hide();
    }
}

$(".buttonWithDiv").click(function() {
    // alert($(this).parent().next());
    nextDiv = $(this).next("div");
    toggleHideShow(nextDiv);
});

$(".hiddenDiv").click(function() {
    toggleHideShow($(this));
});

</script>
    """
    return Str

def css() -> str:
    Str = """
<style>

table, tr, td, th {
    border: 1px solid black;
    border-collapse: collapse;
}

.hiddenDiv {
    display: none;
}

</style>
"""
    return Str

def htmlFormatter(status: Dict, getLoss: bool=True, getBestSamples: bool=True, getLastSamples: bool=True) -> str:
    Str = f"""
<html>
    <head> {css()} </head>
    <h1> NGCDir status </h1>
    <div> Path: {status["Path"]} </div>
    <div> NGC Model type: {status["Cfg"]["graphConfig"]} </div>
"""

    for iteration in range(1, len(status["iterationInfo"]) + 1):
        Str += f"<h2> Iteration {iteration} </h2>"
        Str += getHtmlForData(status, iteration=iteration)
        Str += getHtmlForModels(status, iteration=iteration, getLoss=getLoss, \
            getBestSamples=getBestSamples, getLastSamples=getLastSamples)

    Str += f"""
    <footer> {js()} </footer>
</html>
"""
    return Str



from pathlib import Path
from typing import Dict
from natsort import natsorted
from functools import partial
from media_processing_lib.image import CollageMaker
from .node_plot_functions import buildNodePlotFn
from .utils import loadNpz
from ..nodes import *

class NGCCollageMaker(CollageMaker):
    def __init__(self, baseDir:Path, graphCfg:Dict, outputDir:Path):
        self.baseDir = Path(baseDir)
        self.graphCfg = graphCfg
        files, plotFns, names = self.setup()
        super().__init__(files=files, plotFns=plotFns, outputDir=outputDir, loadFns=loadNpz)

    def setup(self):
        files, names, plotFns = [], [], []
        for name, type in zip(self.graphCfg["nodeNames"], self.graphCfg["nodeTypes"]):
            nodeDir = self.baseDir / name
            assert nodeDir.exists(), f"Not found: '{nodeDir}'"
            nodeFiles = natsorted([str(x.absolute()) for x in nodeDir.glob("*.npz")])
            files.append(nodeFiles)

            # Hack a fake class so we can isntantiate the abstract classes
            Type = globals()[type]
            class _Node(Type):
                def getEncoder():
                    pass
                def getDecoder():
                    pass

            nodeParams = {} if not name \
                in self.graphCfg["hyperParameters"] else self.graphCfg["hyperParameters"][name]
            node = _Node(**nodeParams)
            names.append(name)
            plotFn = buildNodePlotFn(node)
            plotFns.append(plotFn)
        return files, plotFns, names



from typing import List, Dict
from overrides import overrides
from nwgraph import Edge, Node
from ..edges import SingleLink, TwoHopsResidualNoGT, SingleLinkResidual
from ..ngc_v1 import NGCV1
from ...logger import logger

class NGCV2Residual(NGCV1):
    """
    NGCV2 is an NGCV1 with the following changes:
        - there is a single residual node in the cfg, that is also an input node
        - all edges starting with the residual node, are regular single links
        - all edges starting with another input node (not residual), are single link residual edges
        - all two hops are two hops residuals no gt
        - there is no two hop no gt (non-residual)
    """
    def __init__(self, nodes: List[Node], cfg: Dict):
        assert "residualNode" in cfg
        assert cfg["residualNode"] in cfg["inputNodes"]
        super().__init__(nodes, cfg)

    @overrides
    def _getNGCEdgesFromList(self, edgesList: List[List[str]]) -> List[Edge]:
        logger.debug(f"Creating edges from a list of edges ({len(edgesList)}).")
        edges = []
        gtKeyToNode = {node.name: node for node in self.nodes}
        # Only one residual node allowed in this implementation. Both for SLR and THR.
        residualNode = gtKeyToNode[self.cfg["residualNode"]]
        singleLinks = {}

        # Pass once to create single links (residual and not)
        for edge in edgesList:
            edgeNodes = [gtKeyToNode[nodeStr] for nodeStr in edge]
            if len(edge) == 2:
                key = (edgeNodes[0], edgeNodes[1])
                assert key not in singleLinks
                if edge[0] == self.cfg["residualNode"]:
                    edge = SingleLink(edgeNodes[0], edgeNodes[1])
                else:
                    edge = SingleLinkResidual(residualNode, edgeNodes[0], edgeNodes[1])
                singleLinks[key] = edge
            elif len(edge) == 3:
                continue
            else:
                assert False
            edges.append(edge)

        # Pass again to create two hops residual
        for edge in edgesList:
            edgeNodes = [gtKeyToNode[nodeStr] for nodeStr in edge]
            if len(edge) == 2:
                continue
            elif len(edge) == 3:
                key = (edgeNodes[0], edgeNodes[1])
                assert key in singleLinks
                edge = TwoHopsResidualNoGT(singleLinks[key], residualNode, edgeNodes[1], edgeNodes[2])
            else:
                assert False
            edges.append(edge)

        return edges



from typing import Dict, List, Callable
import numpy as np
from nwgraph import Node
from ...logger import logger

# Receive a list of arrays (all nodes) and return same list after augmentation without knowing which nodes are what
GeneralPrototype = Callable[ [List[np.ndarray]], List[np.ndarray] ]
# Node specific augmentation. Receive the node itself (for hyperparameters) and the data, return the augmented data
NodeSpecificPrototype = Callable[ [Node, np.ndarray], np.ndarray ]

class NGCAugmentation:
    # @brief General Augmentation module
    # @param[in] nodes The list of all relevant nodes. Data should assume that the received list is using these nodes.
    # @param[in] general_augmentation A list of ordered augmentation callable functions, that are going to be
    #  invariant of all nodes. All nodes are assumed the same (so 2d only should be passed here).
    # @param[in] node_specific_augmentation A dict of node specific augmentation callables. These are called after
    #  the generic ones.
    def __init__(self, nodes: List[Node], general_augmentation: List[GeneralPrototype], \
                 node_specific_augmentation: Dict[str, List[NodeSpecificPrototype]]=None):
        self.node_name_to_node = {node.name: node for node in nodes}
        if node_specific_augmentation is None:
            node_specific_augmentation = {}
        if general_augmentation is None:
            general_augmentation = []

        self.nodes = nodes
        self.general_augmentation = general_augmentation
        self.node_specific_augmentation = node_specific_augmentation
        logger.debug(f"Instantiating {self}")

    def __call__(self, data: Dict[str, np.ndarray]):
        names, values = data.keys(), list(data.values())
        """Call the generic augmentation callbacks for all data at once."""
        for general_fn in self.general_augmentation:
            values = general_fn(values)
        
        nodes = {self.node_name_to_node[name] for name in names}
        result = dict(zip(names, values))

        """Call the node specific augmentation callbacks for each node individually."""
        for name, node_fns in self.node_specific_augmentation:
            node = nodes[name]
            node_value = result[name]
            for node_fn in node_fns:
                node_value = node_fn(node, node_value)
            result[name] = node_value

        return result

    def __str__(self) -> str:
        count_node_specific = {k: len(self.node_specific_augmentation[k]) for k in self.node_specific_augmentation}
        Str = "[Augmentation]"
        Str += f"\n - Nodes: {self.nodes}"
        Str += f"\n - General augmentation functions: {len(self.general_augmentation)}"
        Str += f"\n - Node specific functions: {count_node_specific}"
        return Str



from functools import lru_cache
from typing import List, Tuple, Any, Dict
from collections import OrderedDict 
import numpy as np
from pathlib import Path
from nwgraph import Node
from ..logger import drange, logger
from ..nodes import *


@lru_cache
def loadNpz(path: Path) -> np.ndarray:
    """Loads a NPZ given a path. Caveats for VRE exported npz."""
    try:
        item = np.load(path, allow_pickle=True)["arr_0"]
    except Exception as e:
        print(f"Error at reading file '{path}'.")
        raise Exception(e)
    # For items exported using VRE.
    if item.dtype == object:
        item = item.item()
        assert isinstance(item, dict)
        assert "data" in item
        item = item["data"]
    assert item.dtype in (np.uint8, np.uint32, np.float16, np.float32, np.float64), f"Got {item.dtype} for {path}"
    if item.shape[0] == 1:
        item = item[0]
    if item.shape[-1] == 1:
        item = item[..., 0]
    return item

def loadNpzFromList(paths: List[Path]) -> np.ndarray:
    """Loads a list of npz given a list of paths using loadNpz function."""
    items = []
    for i in drange(len(paths), desc="Loading Npz"):
        item = loadNpz(paths[i])
        items.append(item)
    items = np.array(items).astype(np.float32)
    return items

def generateRandomData(baseDir: Path, representations: List[str], dims: List[int], types: List[str], \
        resolution: Tuple[int, int], N: int, prefix: str=""):
    """Generates a dataset in ngcdir format"""
    baseDir.mkdir(exist_ok=False, parents=True)
    for i in drange(len(representations)):
        representation = representations[i]
        H, W, D = resolution[0], resolution[1], dims[i]
        (baseDir / representation).mkdir()
        if types[i] == "float":
            data = np.random.rand(N, H, W, D).astype(np.float32)
        elif types[i] == "categorical":
            data = np.random.randint(0, D-1, size=(N, H, W)).astype(np.uint32)
        else:
            assert False
        for j in drange(N, desc=representation):
            outFile = baseDir / representation / (f"{prefix}{j}.npz")
            np.savez_compressed(outFile, data[j])

def getFakeNodeType(nodeType: type) -> type:
    """Given a non-instantaible node from this library, create a fake node that implements the getEncoder and
    getDecoder methods.
    """
    class FakeNode(nodeType):
        def getEncoder(self, outputNode):
            pass
        def getDecoder(self, inputNode):
            pass
    return FakeNode

def instantiateNodes(names: List[str], types: List[str],
        nodesHyperparameters: Dict[str, Dict[str, Any]]={}) -> List[Node]:
    """Instantiates a list of nodes, given names and types and, optionally, hyperparameters"""

    res = OrderedDict()
    resList = []
    for name, type in zip(names, types):
        nodeType = globals()[type]
        nodeParams = {} if (nodesHyperparameters is None or name not in nodesHyperparameters.keys()) \
            else nodesHyperparameters[name]
        if name in res:
            logger.debug(f"Node '{name}' already instantiated! Copying.")
            node = res[name]
            assert isinstance(node, nodeType), "Instantaited node has different type."
        else:
            try:
                node = nodeType(**nodeParams)
            except Exception:
                logger.debug(f"Cannot instantaite '{name}' (type: {nodeType}). "
                    "Creating a fake node and trying again.")
                node = getFakeNodeType(nodeType)(**nodeParams)
        res[name] = node
        resList.append(node)
    return resList



import imp


import numpy as np

class DeterministicAugmentation:
    def __init__(self, seed: int, num_precomputted: int=100):
        assert isinstance(seed, int)
        self.seed = seed
        np.random.seed(seed)
        self.next_seeds = np.random.randint(0, 1000, size=(num_precomputted, ))
        self.index = 0

    def __call__(self):
        next_seed = self.next_seeds[self.index % len(self.next_seeds)]
        np.random.seed(next_seed)
        self.index += 1



from typing import Dict
from overrides import overrides
from torchmetrics import Metric

from .map_node import MapNode
from .rgb import RGB

class Halftone(RGB):
	def __init__(self, name:str="halftone"):
		MapNode.__init__(self, name=name, numDims=3)

	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		metrics = super().getNodeMetrics()
		newMetrics = {}
		for metricName, metricFn in metrics.items():
			newMetricName = metricName.replace("RGB", "Halftone")
			newMetrics[newMetricName] = metricFn
		return newMetrics


import numpy as np
import torch as tr
import torch.nn.functional as F
from typing import Dict
from collections import OrderedDict
from sklearn.metrics import f1_score, jaccard_score
from torchmetrics.functional import accuracy
from nwgraph import Node
from nwutils.torch import npToTrCall

from ..nodes import Semantic

# For a given node with a set of predictions and GT, return all the node's metrics.
# Some hardcoded stuff for Semantic, until we figure out if we can generalize it better.
def getMetrics(y:np.ndarray, t:np.ndarray, node:Node) -> Dict:
    if isinstance(node, Semantic):
        if y.shape[-1] == node.numClasses:
            argMaxY = np.argmax(y, -1)
        else:
            argMaxY = y
        argMaxY = argMaxY.astype(np.uint8).flatten()
        oneHotY = F.one_hot(tr.from_numpy(argMaxY).type(tr.int64), node.numClasses).type(tr.float32)

        if t.shape[-1] == node.numClasses:
            argMaxT = np.argmax(t, -1).astype(np.uint8)
        else:
            argMaxT = t.astype(np.uint8)
        argMaxT = argMaxT.astype(np.uint8).flatten()
        oneHotT = F.one_hot(tr.from_numpy(argMaxT).type(tr.int64), node.numClasses).type(tr.float32)

        Loss = node.getNodeCriterion()(oneHotY, oneHotT)
        metrics = {
            "Accuracy": lambda y, t: float(accuracy(tr.from_numpy(argMaxY), tr.from_numpy(argMaxT))),
			"Mean IoU": lambda y, t : jaccard_score(argMaxY, argMaxT, average=None, labels=range(node.numClasses)),
			"F1 Score": lambda y, t : f1_score(argMaxY, argMaxT, average=None, labels=range(node.numClasses)),
        }
    else:
        Loss = npToTrCall(node.getNodeCriterion(), y, t)
        metrics = node.getNodeMetrics()
    res = {k:metrics[k](y, t) for k in metrics.keys()}
    newRes = OrderedDict()
    newRes["Loss"] = float(Loss)
    for k in res:
        if isinstance(res[k], np.ndarray):
            newRes[k] = list(res[k])
            newRes[f"{k} (mean)"] = res[k].mean()
        else:
            newRes[k] = res[k]
    return newRes



from typing import Dict

def getDataStr(dataStatus: Dict) -> str:
    """Parses the data status of an iteration."""
    Str = "[Data]"
    if dataStatus == "n/a":
        Str += "\n  - n/a"
        return Str
    for node in dataStatus.keys():
        strTypes = ", ".join(dataStatus[node]["typesSemisupervised"])
        Str += f"\n  - Node: {node}"
        Str += f"\n    - Total={dataStatus[node]['numAll']}. Train={dataStatus[node]['numTrain']} "
        Str += f"Semisupervised={dataStatus[node]['numSemisupervised']} ({strTypes})"
    return Str

def getModelStr(modelStatus: Dict) -> str:
    """Parses the model status of an iteration"""
    Str = "[Model]"
    Str += f"\n  - Total edges: {modelStatus['totalEdges']}"
    Str += f"\n  - Total trained: {modelStatus['totalTrained']}"
    Str += f"\n  - Edges:"
    for edge, values in modelStatus["edges"].items():
        Str += f"\n    - {edge} -- "
        Str += f"Total epochs: {values['trainedEpochs']}. "
        Str += f"Best epoch: {values['bestEpoch']}. "
        Str += f"Duration: {values['duration']['total']}."
    Str += f"\n  - Total duration: {modelStatus['totalDuration']}"
    return Str

def strFormatter(status: Dict) -> str:
    """Given a status obtained via NgcDir(Path).getStatus(), returns a formatted string for pretty printing."""
    Str = "[NGCDir]"
    Str += f"\nPath: {status['Path']}"

    # graph cfg information (nodes, edges, hyperparameters, etc.)
    Str += "\nCfg: "
    for k in status["Cfg"].keys():
        Str += f"\n  - {k}: {status['Cfg'][k]}"

    # Iteration results (data and edges)
    Str += f"\nAvailable iterations: {status['numIterations']}"
    for iteration in range(1, status["numIterations"] + 1):
        iterDataStr = getDataStr(status["iterationInfo"][iteration]["data"])
        iterModelStr = getModelStr(status["iterationInfo"][iteration]["model"])
        Str += f"\n\n-- Iteration {iteration} --"
        Str += f"\n{iterDataStr}"
        Str += f"\n\n{iterModelStr}"
        Str += "\n________________________________________________________"
    return Str



from typing import Dict
from functools import partial
import numpy as np
from pathlib import Path
from nwgraph import Edge
from nwmodule.trainer import setup_module_for_train
from nwmodule.callbacks import RandomPlotEachEpoch
from media_processing_lib.image import image_write
from nwutils.torch import trGetData, trToDevice, npGetData

from ..models.edges import TwoHopsNoGT, TwoHopsResidualNoGT, SingleLink, SingleLinkResidual
from ..utils import buildNodePlotFn

def edge_plot_fn(x: np.ndarray, y: np.ndarray, t: np.ndarray, workingDirectory: Path, edge: Edge):
    cnt = 0
    MB = len(t)
    A, B = edge.getNodes()
    if isinstance(edge, (SingleLink, SingleLinkResidual)):
        """For single links, we have access to the input data, so we plot that."""
        x = x[edge.inputNode.name]
    elif isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)):
        """For TwoHops, we need to propagate through the single link first, so we get intermediate the prediction."""
        x = trToDevice(trGetData(x), edge.getDevice())
        x = edge.singleLinkEdge.networkAlgorithm(x, None, False, False)[0]
    else:
        assert False, f"Unknown edge: {edge}."

    x, y, t = npGetData([x, y, t])
    for i in range(MB):
        xImage = buildNodePlotFn(A)(x[i])
        yImage = buildNodePlotFn(B)(y[i])
        tImage = buildNodePlotFn(B)(t[i])

        stack = np.concatenate([xImage, yImage, tImage], axis=1)
        image_write(stack, f"{workingDirectory}/{cnt}.png")
        cnt += 1

def setup_edge(edge: Edge, trainCfg: Dict, weightsDir: Path):
    setup_module_for_train(edge, trainCfg)
    if isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)):
        from .edge_trainer import EdgeTrainer
        edge.setup(singleLinkDir = weightsDir / EdgeTrainer.getEdgeDirName(edge.singleLinkEdge))
    assert edge.getNumParams()[1] > 0, f"Edge '{edge}' has no trainable params."

    edge.addCallback(RandomPlotEachEpoch(partial(edge_plot_fn, edge=edge)))



import torch as tr
import numpy as np
from nwmodule import CriterionType
from torchmetrics import Metric
from typing import Dict
from overrides import overrides
from functools import partial

from .map_node import MapNode

class Depth(MapNode):
	def __init__(self, maxDepthMeters:float, name:str="depth"):
		self.maxDepthMeters = maxDepthMeters
		hyperParameters = {"maxDepthMeters":maxDepthMeters}
		super().__init__(name = name, numDims = Depth.numDims(), hyperParameters = hyperParameters)

	@staticmethod
	def numDims() -> int:
		return 1

	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		metrics = {
			"Depth (m)" : partial(Depth.depthMetric, maxDepthMeters=self.maxDepthMeters, gtKey=self.name),
			"RMSE" : partial(Depth.rmseMetric, maxDepthMeters=self.maxDepthMeters, gtKey=self.name)
		}
		return metrics

	@overrides
	def getNodeCriterion(self) -> CriterionType:
		return lambda y, t: Depth.lossFn(y, t)

	@staticmethod
	def lossFn(y: tr.Tensor, t: tr.Tensor):
		assert y.shape == t.shape
		L = ((y - t)**2).mean()
		return L

	def depthMetric(y: np.ndarray, t: np.ndarray, maxDepthMeters: float, **k) -> float:
		# Normalize back to meters, output is in [0 : 1] representing [0 : maxDepthMeters]m
		yDepthMeters = y * maxDepthMeters
		tDepthMeters = t * maxDepthMeters
		l1 = np.abs(yDepthMeters - tDepthMeters).mean()
		return l1

	def rmseMetric(y: np.ndarray, t: np.ndarray, maxDepthMeters: float, **k) -> float:
		# Normalize back to milimeters, output is in [0 : 1] representing [0 : maxDepthMeters]m
		yDepthMeters = y * maxDepthMeters
		tDepthMeters = t * maxDepthMeters
		L2 = (yDepthMeters - tDepthMeters) ** 2
		rmse = np.sqrt(L2.mean())
		return rmse



from pathlib import Path
from typing import Dict
from overrides import overrides
from nwmodule.trainer import NWTrainer
from .edge_trainer import EdgeTrainer, get_loader
from .setup_edge import setup_edge
from ..readers import EdgeReader, NGCNpzReader
from ..readers.augmentation import get_augmentation
from ..logger import logger

class EdgeTrainerSequential(EdgeTrainer):
    @overrides
    def run(self, trainCfg: Dict, trainReader: NGCNpzReader, validationReader: NGCNpzReader, debug: bool):
        if debug:
            trainCfg["numEpochs"] = 3
        weightsDir = Path.cwd()

        trained = self.model.getTrainedEdges(weightsDir, EdgeTrainer.getEdgeDirName)
        for edge in self.model.edges:
            logger.debug(f"Current edge: {edge}")
            setup_edge(edge, trainCfg, weightsDir)
            if edge in trained:
                logger.debug(f"Edge '{edge}' already trained. Skipping.")
                continue

            trainLoader = get_loader(trainReader, edge, trainCfg, randomize=True, debug=debug, augmentation=True)
            valLoader = get_loader(validationReader, edge, trainCfg, randomize=False, debug=debug, augmentation=False)

            Dir = Path.cwd() / EdgeTrainer.getEdgeDirName(edge)
            trainer = NWTrainer(edge, Dir)
            trainer.train(trainLoader, trainCfg["numEpochs"], valLoader)
            del valLoader
            del trainLoader



import torch as tr
from typing import Dict, List
from .ngc import NGC
from .ngc_v1 import NGCAAAI2020, NGCV1
from .ngc_v2_residual import NGCV2Residual
from nwgraph import Node

device = tr.device("cuda") if tr.cuda.is_available() else tr.device("cpu")

def getModel(nodes:List[Node], cfg:Dict) -> NGC:
	modelType = cfg["graphConfig"]
	if modelType == "NGC_AAAI2020":
		model = NGCAAAI2020(nodes=nodes, cfg=cfg)
	elif modelType == "NGC-V1":
		model = NGCV1(nodes=nodes, cfg=cfg)
	elif modelType == "NGC-V2-Residual":
		model = NGCV2Residual(nodes=nodes, cfg=cfg)
	else:
		assert False, f"Unknown model: {modelType}"
	model.setTrainableWeights(True)
	model = model.to(device)
	return model



from .map_node import MapNode
from .vector_node import VectorNode

# Other nodes
from .dummy_identity_map_node import DummyIdentityMapNode
from .null_node import NullNode

# Custom nodes go here
from .rgb import RGB
from .hsv import HSV
from .depth import Depth
from .edges import Edges
from .normal import Normal
from .cameranormal import CameraNormal
from .semantic import Semantic
from .halftone import Halftone
from .wireframe import Wireframe
from .softsegmentation import SoftSegmentation
from .pose import Pose


import numpy as np
import torch
import torch.nn.functional as F
from nwmodule import CriterionType
from torchmetrics import Metric
from typing import Dict
from overrides import overrides

from .map_node import MapNode

class Normal(MapNode):
	def __init__(self, name: str = "normal"):
		super().__init__(name = name, numDims = 3)

	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		return {
			"Normal-v1 (deg)": Normal.degreeMetricV1,
			"Normal-v2 (deg)": Normal.degreeMetricV2,
			"Normal (deg)": Normal.degreeMetric,
		}

	@overrides
	def getNodeCriterion(self) -> CriterionType:
		return Normal.lossFn

	@staticmethod
	def degreeMetricV1(y, t, **k):
		# First, remove sign from both and then do the L1 diff
		y = np.abs(y)
		t = np.abs(t)
		diff = np.abs(y - t)
		return diff.mean() * 360

	@staticmethod
	def degreeMetricV2(y, t, **k):
		# First, remove sign from both and then do the L1 diff
		y = np.abs(y)
		t = np.abs(t)
		y = torch.from_numpy(y)
		t = torch.from_numpy(t)
		cosine_distance = F.cosine_similarity(y, t)
		cosine_distance = np.minimum(np.maximum(cosine_distance.cpu().numpy(), -1.0), 1.0)
		angles = np.arccos(cosine_distance) / np.pi * 180.0
		return angles.mean()

	@staticmethod
	def degreeMetric(y, t, **k):
		# First, remove sign from both and then do the L1 diff
		y = np.abs(y)
		t = np.abs(t)
		y = torch.from_numpy(y)
		t = torch.from_numpy(t)
		cosine_distance = F.cosine_similarity(y, t, dim=-1)
		cosine_distance = cosine_distance.cpu().numpy()
		cosine_distance = np.abs(cosine_distance)
		angles = np.arccos(cosine_distance) / np.pi * 180
		angles[np.isnan(angles)] = 180
		return angles.mean()

	@staticmethod
	def lossFn(y, t):
		return ((y - t)**2).mean()


from .ngc_edge import NGCEdge
from .single_link import SingleLink
from .two_hops_nogt import TwoHopsNoGT
from .single_link_residual import SingleLinkResidual
from .two_hops_residual_nogt import TwoHopsResidualNoGT



# TODOs
# - automatic graph detection based on performance

import os
from typing import List, Dict, Union
from pathlib import Path

from .edge_trainer import EdgeTrainer
from .edge_trainer_sequential import EdgeTrainerSequential
from .edge_trainer_parallel import EdgeTrainerParallel
from .ngc_iteration_prologue import iterationPrologue
from ..models import NGC
from ..logger import logger
from ..readers import NGCNpzReader
from ..ngcdir import NGCDir
from ..dir_linker import DirLinker

def changeDirectory(Dir:Union[str, Path]):
	Dir = Path(Dir)
	Dir.mkdir(exist_ok=True, parents=True)
	logger.debug(f"Changing to working directory: {Dir}")
	os.chdir(Dir)

class NGCTrainer:
	def __init__(self, model: NGC, ngcDirPath: Path, method: str, trainCfg: Dict,
		trainDir: Path, validationDir: Path, semisupervisedDirs: List[Path]):
		self.model = model
		ngcDirPath.mkdir(exist_ok=True, parents=True)
		self.ngcDir = NGCDir(ngcDirPath, model.cfg)
		self.dirLinker = DirLinker(self.ngcDir, trainDir, validationDir, semisupervisedDirs)
		self.method = method
		self.trainCfg = trainCfg
		self.nameToNode = {node.name: node for node in self.model.getNodes()}

	# @brief Training for 1 iteration only
	def doOneIteration(self, edgeTrainer: EdgeTrainer, iteration:int, debug:bool):
		logger.info(f"Starting iteration {iteration}")
		modelDir = self.ngcDir.getAllDataDirs(iteration)["models"][iteration]
		dataDir = self.ngcDir.getAllDataDirs(iteration)["data"][iteration]
		cwd = Path.cwd()
		changeDirectory(modelDir)

		inputNodes = [self.nameToNode[x] for x in self.model.cfg["inputNodes"]]
		trainReader = NGCNpzReader(path=dataDir, nodes=self.model.getNodes(), gtNodes=inputNodes)
		validationReader = NGCNpzReader(path=self.dirLinker.validationDir, \
			nodes=self.model.getNodes(), gtNodes=inputNodes)

		self.model.train()
		edgeTrainer.run(self.trainCfg, trainReader, validationReader, debug)
		changeDirectory(cwd)

	# @brief The main public function to be called on an NGC model
	def run(self, startIteration: int, endIteration: int, parallelization: bool = False, debug: bool = False):
		logger.info(f"Training the NGC graph from iteration {startIteration} to {endIteration} " + \
				f"(parallelization: {parallelization}, debug: {debug}).")
		self.dirLinker.linkTrainDir(iteration=1)

		if parallelization:
			edgeTrainer = EdgeTrainerParallel(self.model)
		else:
			edgeTrainer = EdgeTrainerSequential(self.model)

		for iteration in range(startIteration, endIteration + 1):
			self.model.train()
			self.doOneIteration(edgeTrainer, iteration, debug)

			self.model.eval()
			iterationPrologue(self, iteration, debug)



import numpy as np
import torch as tr
from nwgraph import CriterionType
from torchmetrics import Metric
from typing import Dict
from overrides import overrides

from .vector_node import VectorNode

class PositionMetric(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_state("result", default=tr.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=tr.tensor(0), dist_reduce_fx="sum")

    def update(self, y: tr.Tensor, gt: tr.Tensor):
        assert y.shape == gt.shape
        y, gt = y[:, 0:3], gt[:, 0:3]
        error = tr.norm(y - gt, dim=-1)
        self.result += error
        self.total += y.numel()

    def compute(self):
        return self.result.float() / self.total

class OrientationMetric(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_state("result", default=tr.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=tr.tensor(0), dist_reduce_fx="sum")

    def update(self, y: tr.Tensor, gt: tr.Tensor):
        assert y.shape == gt.shape
        y, gt = y[:, 3:], gt[:, 3:]

        # [0 : 1] => [-1 : 1]
        y = y * 2 - 1
        gt = gt * 2 - 1

        # https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
        # [-1 : 1] => [0 : 1]
        error = tr.abs((y - gt + 1) % 2 - 1)
        # error :: [0 : 1] => [0 : 180]
        error = error * 180
        self.result += error
        self.total += y.numel()

    def compute(self):
        return self.result.float() / self.total

class Pose(VectorNode):
    def __init__(self, name: str="pose"):
        super().__init__(name=name, numDims=6)

    @overrides
    def getNodeMetrics(self) -> Dict[str, Metric]:
        return {
            "Position": PositionMetric(),
            "Orientation (deg)": OrientationMetric()
        }

    @overrides
    def getNodeCriterion(self) -> CriterionType:
        return Pose.lossFn

    def lossFn(y, t):
        return ((y - t)**2).mean()



"""Python logger settings."""

import os
import logging
from tqdm import trange

KEY = "NGCLIB"
ENV_KEY = f"{KEY}_LOGLEVEL"
env_var = int(os.environ[ENV_KEY]) if ENV_KEY in os.environ else 2

# Usage: loglevel=0 (none), loglevel=1 (info), loglevel=2 (debug), loglevel=3 (debug verbose)
logging.DEBUG2 = 3
loglevel = {0: logging.NOTSET, 1: logging.INFO, 2: logging.DEBUG, 3: logging.DEBUG2}[env_var]
logging.addLevelName(logging.DEBUG2, "DEBUG-VERBOSE")


class drange:
    """Use `drange` instead of `range` and `tqdm.trange`.
    If {Key}_TQDM is set to 1, `drange` will call `trange`, otherwise will call `range`.
    """

    def __init__(self, *args, **kwargs):
        tqdm_key = f"{KEY}_TQDM"
        self.env_var = bool(int(os.environ[tqdm_key])) if tqdm_key in os.environ else True
        self.range = trange(*args, **kwargs) if self.env_var else range(*args)

    def __iter__(self):
        return self.range.__iter__()

    # def __next__(self):
    #     return self.range.__next__()

    def set_description(self, *args, **kwargs):
        """Set description."""
        if self.env_var:
            self.range.set_description(*args, **kwargs)


class CustomFormatter(logging.Formatter):
    """Custom formatting for logger."""

    yellow = "\x1b[33;20m"
    green = "\x1b[32;20m"
    cyan = "\x1b[36;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    pre = "[%(asctime)s-%(name)s-%(levelname)s]"
    post = "(%(filename)s:%(funcName)s:%(lineno)d)"

    # Example [TIME:LEVEL:NAME] Message [FILE:FUNC:LINE]
    FORMATS = {
        logging.DEBUG: f"{cyan}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.DEBUG2: f"{cyan}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.INFO: f"{green}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.WARNING: f"{yellow}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.ERROR: f"{red}{pre}{reset} %(message)s {yellow}{post}{reset}",
        logging.CRITICAL: f"{bold_red}{pre}{reset} %(message)s {yellow}{post}{reset}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        formatter.formatTime = self.formatTime
        return formatter.format(record)

    def formatTime(self, record, datefmt=None):
        return super().formatTime(record, "%Y-%m-%d %H:%M:%S")

# instantiate logger and set log level
logger = logging.getLogger(KEY)
logger.setLevel(loglevel)
logger.debug2 = lambda msg: logger.log(logging.DEBUG2, msg)

# add custom formatter to logger
handler = logging.StreamHandler()
handler.setFormatter(CustomFormatter())
logger.addHandler(handler)



import numpy as np
from nwmodule import CriterionType
from torchmetrics import Metric
from typing import Dict
from overrides import overrides

from .map_node import MapNode

class RGB(MapNode):
	def __init__(self, name: str = "rgb"):
		super().__init__(name = name, numDims = RGB.numDims())

	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		return {
			"RGB (L1 pixel)" : RGB.RGBMetricL1Pixel,
			"RGB (L2)" : RGB.RGBMetricL2
		}

	@overrides
	def getNodeCriterion(self) -> CriterionType:
		return RGB.lossFn

	@staticmethod
	def numDims() -> int:
		return 3

	@staticmethod
	def lossFn(y, t):
		return ((y - t)**2).mean()

	@staticmethod
	def RGBMetricL1Pixel(y, t, **k):
		# Remap y and t from [0 : 1] to [0 : 255]
		yRgbOrig = y * 255
		tRgbOrig = t * 255
		return np.abs(yRgbOrig - tRgbOrig).mean()

	@staticmethod
	def RGBMetricL2(y, t, **k):
		return ((y - t)**2).mean()



from typing import Dict, List, Union, Tuple
from functools import partial
from overrides import overrides
from torchmetrics import Metric
from torchmetrics.functional import accuracy
from nwmodule.nwmodule import CriterionType
from nwmodule.loss import softmax_nll

from .map_node import MapNode

def dice_loss(y, t, nClasses: int, classWeights: List=None):
	if classWeights is None:
		classWeights = [1] * nClasses

	smooth = 1.
	loss = 0.
	for c in range(nClasses):
		iflat = y[..., c].view(-1)
		tflat = t[..., c].view(-1)
		intersection = (iflat * tflat).sum()
		
		w = classWeights[c]
		up = (2. * intersection + smooth)
		down = (iflat.sum() + tflat.sum() + smooth)
		loss += w * (1 - up / down)
	return loss

class Semantic(MapNode):
	def __init__(self, semanticClasses: Union[int, List[str]], semanticColors: List[Tuple[int, int, int]], \
			name: str="semantic", useGlobalMetrics: bool=False, criterionType: str="cross_entropy"):

		if isinstance(semanticClasses, int):
			from ..utils import logger
			logger.debug("Semantic classes were not provided, just the number. Creating fake classes.")
			semanticClasses = list(range(semanticClasses))
		assert isinstance(semanticClasses, (list, tuple))

		super().__init__(name=name, numDims=len(semanticClasses))
		self.semanticClasses = semanticClasses
		self.semanticColors = semanticColors
		self.numClasses = len(semanticClasses)
		self.useGlobalMetrics = useGlobalMetrics
		self.criterionType = criterionType
		assert len(self.semanticClasses) == len(self.semanticColors), f"{self.semanticClasses} " + \
			f"({len(self.semanticClasses)}) vs {self.semanticColors} ({len(self.semanticColors)})"

	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		metrics = {
			"Accuracy" : lambda y, t: accuracy(y.argmax(-1), t.argmax(-1)),
		}
		assert not self.useGlobalMetrics
		return metrics

	@overrides
	def getNodeCriterion(self) -> CriterionType:
		criterion = {
			"cross_entropy": Semantic.cross_entropy,
			"cross_entropy_plus_dice": partial(Semantic.cross_entropy_plus_dice, nClasses=self.numClasses),
		}[self.criterionType]
		return criterion

	@staticmethod
	def cross_entropy(y, t):
		return softmax_nll(y, t, dim=-1).mean()

	@staticmethod
	def cross_entropy_plus_dice(y, t, nClasses):
		cross_entropy = softmax_nll(y, t, dim=-1).mean()
		dice = dice_loss(y, t, nClasses)
		return cross_entropy + dice



import numpy as np
from typing import Dict
from overrides import overrides
from torchmetrics.functional import accuracy
from torchmetrics import Metric
from nwmodule import CriterionType
from nwmodule.loss import sigmoid_bce

from .rgb import RGB
from .map_node import MapNode

class Wireframe(RGB):
	def __init__(self, name:str="wireframe", useGlobalMetrics:bool=False):
		MapNode.__init__(self, name=name, numDims=1)
		self.useGlobalMetrics = useGlobalMetrics
	
	@overrides
	def getNodeMetrics(self) -> Dict[str, Metric]:
		metrics = {
			"Accuracy" : lambda y, t: accuracy(y.argmax(-1), t.argmax(-1)),
		}
		if self.useGlobalMetrics:
			# TODO: These return 0s?
			# metrics["Mean IoU (global)"] = MeanIoU(mode="global")
			# metrics["F1 Score (global)"] = F1Score(mode="global", returnMean=False)
			metrics["Accuracy (global)"] = Accuracy(mode="global")
		return metrics

	@overrides
	def getNodeCriterion(self) -> CriterionType:
		return Wireframe.lossFn

	def lossFn(y:np.ndarray, t:np.ndarray) -> float:
		return sigmoid_bce(y[..., 0], t[..., 0]).mean()



import time
import threading
import torch as tr
from torch.utils.data import DataLoader
from queue import Queue
from overrides import overrides
from typing import Dict, List
from pathlib import Path
from nwgraph import Edge
from nwmodule.trainer import NWTrainer

from .edge_trainer import EdgeTrainer, get_loader
from .setup_edge import setup_edge
from ..logger import logger
from ..readers import NGCNpzReader
from ..models import NGC
from ..models.edges import TwoHopsNoGT, TwoHopsResidualNoGT

def trainOneEdge(device: tr.device, q: Queue, edge: Edge, \
        trainCfg: Dict, weightsDir: Path, edgeTrainLoader: DataLoader, edgeValidationLoader: DataLoader):
    logger.info(f"[device:{device.index}] Training edge '{edge}'.")
    try:
        edge.to(device)
        setup_edge(edge, trainCfg, weightsDir)
        Dir = Path.cwd() / EdgeTrainer.getEdgeDirName(edge)
        trainer = NWTrainer(edge, Dir)
        trainer.train(edgeTrainLoader, trainCfg["numEpochs"], edgeValidationLoader)
    except Exception as e:
        logger.debug(f"[device:{device.index}] Error while training edge '{edge}'. Exception: {str(e)}")
        breakpoint()
    logger.debug(f"[device:{device.index}] Finished training edge '{edge}'")
    q.put(device)
    del edgeTrainLoader
    del edgeValidationLoader

def getTorchAvailableGPUs() -> List[tr.device]:
    # Torch takes into account CUDA_VISIBLE_DEVICES and also reorders ids starting from 0, ignoring global cuda id
    torchAvailableDevices = []
    i = 0
    while True:
        try:
            device = tr.device(f"cuda:{i}")
            _ = tr.FloatTensor([0]).to(device)
            torchAvailableDevices.append(device)
        except Exception:
            break
        i += 1
    logger.debug(f"Torch available devices: {len(torchAvailableDevices)}")
    assert len(torchAvailableDevices) > 0
    return torchAvailableDevices

class EdgeTrainerParallel(EdgeTrainer):
    def __init__(self, model: NGC):
        self.model = model
        # This holds the torch-level available gpu ids
        self.torchAvailableDevices = getTorchAvailableGPUs()
        logger.debug(f"Num GPUs available: {len(self.torchAvailableDevices)}")

    @overrides
    def run(self, trainCfg: Dict, trainReader: NGCNpzReader, validationReader: NGCNpzReader, debug: bool):
        # Create a queue variable which will be shared amongst threads to announce when an edge was trained.
        q = Queue()

        if debug:
            trainCfg["numEpochs"] = 3
        weightsDir = Path.cwd()

        trained = self.model.getTrainedEdges(weightsDir, EdgeTrainer.getEdgeDirName)
        untrained, untrainedNames, inProgress, finished = [], [], [], []
        deviceToEdge = {}
        threads = {}
        for i, edge in enumerate(self.model.edges):
            if edge in trained:
                finished.append(edge)
                continue
            untrained.append(edge)
            untrainedNames.append(EdgeTrainer.getEdgeDirName(edge))
        logger.info(f"Num untrained edges: {len(untrained)} / {len(self.model.cfg['edges'])}")
        # Load all available gpus in this main thread
        for i in range(len(self.torchAvailableDevices)):
            q.put(self.torchAvailableDevices[i])

        self.model.to("cpu")
        while True:
            # Busy waiting, since the main thread is not really that relevant
            time.sleep(1)
            if q.empty():
                logger.debug2("Q is empty. Waiting...")
                continue

            device = q.get()
            logger.debug2(f"Got device {device} (id: {device.index})")

            # If this device was previously used by an edge, mark that edge as trained and remove it from untrained. Also
            #  join the thread so resources are unallocated properly
            if device.index in threads:
                assert device.index in deviceToEdge
                edge = deviceToEdge[device.index]
                assert edge in inProgress
                thread = threads.pop(device.index)
                thread.join()
                edge = inProgress.pop(inProgress.index(edge))
                finished.append(edge)
            
            if len(finished) == len(self.model.cfg["edges"]):
                assert len(inProgress) == 0
                assert len(untrained) == 0
                assert len(untrainedNames) == 0
                assert len(threads) == 0
                break
            
            # Still some in progress
            if len(untrained) == 0:
                q.put(device)
                continue

            edge = untrained.pop()
            edgeName = untrainedNames.pop()
            # We must wait for the single link to finish first
            # TODO: Make this generic using topo-sort.
            if isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)):
                if not str(edge.singleLinkEdge) in str(finished):
                    q.put(device)
                    untrained.insert(0, edge)
                    untrainedNames.insert(0, edgeName)
                    continue

            # Otherwise, pop a new edge and start a new thread.
            inProgress.append(edge)
            trainLoader = get_loader(trainReader, edge, trainCfg, randomize=True, debug=debug, augmentation=True)
            valLoader = get_loader(validationReader, edge, trainCfg, randomize=False, debug=debug, augmentation=False)
            threadArgs = (device, q, edge, trainCfg, weightsDir, trainLoader, valLoader)
            threads[device.index] = threading.Thread(target=trainOneEdge, args=threadArgs)
            deviceToEdge[device.index] = edge
            threads[device.index].start()
        self.model.to(self.torchAvailableDevices[0])
        logger.debug("Finished all threads!")



from typing import List, Dict
from copy import copy
from overrides import overrides
from nwgraph import Edge, Message
from ..edges import SingleLink, TwoHopsNoGT, TwoHopsResidualNoGT, SingleLinkResidual
from ..ngc import NGC
from ...logger import logger

class NGCV1(NGC):
    """This class contains the specific quircks of the current NGC implementation, some of which (but not all):
        - Two types of edges: Single Links and Two Hops no GT
        - Two iterations per forward (1st for SL, 2nd for TH)
        - Two types of nodes: Input Nodes and Output Nodes. Only Input Nodes participate in Single Links.
        - TODO: Others when I recall them
    """
    @staticmethod
    def _getNGCEdgesComplete(nodes: List, inputNodes: List) -> List[List[str]]:
        outputNodes = list(set(nodes).difference(inputNodes))
        allEdges = []
        logger.debug(f"Creating complete graph from nodes: {nodes}. Input nodes: {inputNodes}.")

        # Single links
        for inputNode in inputNodes:
            for outputNode in outputNodes:
                allEdges.append([inputNode, outputNode])

        # Two Hops
        for inputNode in inputNodes:
            for i in range(len(outputNodes)):
                for j in range(i + 1, len(outputNodes)):
                    allEdges.append([inputNode, outputNodes[i], outputNodes[j]])
                    allEdges.append([inputNode, outputNodes[j], outputNodes[i]])

        return allEdges

    def _getNGCEdgesFromList(self, edgesList: List[List[str]]) -> List[Edge]:
        logger.debug(f"Creating edges from a list of edges ({len(edgesList)}).")
        edges = []
        gtKeyToNode = {node.name:node for node in self.nodes}
        singleLinks = {}

        # Pass once to create single links
        for edge in edgesList:
            edgeNodes = [gtKeyToNode[nodeStr] for nodeStr in edge]
            if len(edge) == 2:
                edge = SingleLink(edgeNodes[0], edgeNodes[1])
                key = (edgeNodes[0], edgeNodes[1])
                assert key not in singleLinks
                singleLinks[key] = edge
            elif len(edge) == 3:
                continue
            else:
                assert False
            edges.append(edge)

        # Pass again to create two hops
        for edge in edgesList:
            edgeNodes = [gtKeyToNode[nodeStr] for nodeStr in edge]
            if len(edge) == 2:
                continue
            elif len(edge) == 3:
                key = (edgeNodes[0], edgeNodes[1])
                assert key in singleLinks
                edge = TwoHopsNoGT(singleLinks[key], edgeNodes[1], edgeNodes[2])
            else:
                assert False
            edges.append(edge)
        return edges

    @overrides
    def getNGCEdges(self) -> List[Edge]:
        if self.cfg["edges"] == "complete":
            edges = NGCV1._getNGCEdgesComplete(self.cfg["nodeNames"], self.cfg["inputNodes"])
        else:
            edges = self.cfg["edges"]
        assert isinstance(edges, (list, tuple))
        return self._getNGCEdgesFromList(edges)

    def getSingleLinks(self) -> List[Edge]:
        singleLinks = []
        for edge in self.edges:
            if isinstance(edge, (SingleLink, SingleLinkResidual)):
                singleLinks.append(edge)
        return singleLinks

    def getTwoHops(self) -> List[Edge]:
        twoHops = []
        for edge in self.edges:
            if isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)):
                twoHops.append(edge)
        return twoHops

    @overrides
    def messagePass(self, t: int):
        assert 0 <= t <= 1
        def isGTMessage(message):
            # GT messages have the following form: ('GT (rgb)',) or ('GT (rgb)', Single Link (rgb -> softseg2)) etc.
            # Basically tuples of paths. Vote messages will "kill" the pathways (though in future/other graph
            #  implementation could allow gradients to flow through the voting path as well).
            return len(message.path) == 1 and isinstance(message.path[0], str) and message.path[0][0 : 2] == "GT"

        def getSingleLinkMessages(edge: Edge, messages: List[Message]) -> List[Message]:
            return list(filter(isGTMessage, messages))

        def getTwoHopsMessages(edge: Edge, messages: List[Message]) -> List[Message]:
            res = []
            for message in messages:
                if isGTMessage(message):
                    continue
                path = message.path[0]
                # GT (rgb) => ["GT", "rgb"]
                GT, node = path.split(" ")
                node = node[1 : -1]
                assert GT == "GT"
                if node != edge.singleLinkNode.name:
                    continue
                res.append(message)
            assert len(res) > 0
            return res

        # Get all messages before doing any propagation so we store the state
        iterationMessages = {node: copy(node.getMessages()) for node in self.nodes}
        for edge in self.edges:
            if isinstance(edge, (SingleLink, SingleLinkResidual)) and t == 1:
                continue
            if isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)) and t == 0:
                continue

            A, B = edge.getNodes()
            allMessages = iterationMessages[A]
            # for single links, only pass the GT data
            if isinstance(edge, (SingleLink, SingleLinkResidual)):
                validMessages = getSingleLinkMessages(edge, allMessages)
            # for dual hops, only pass the data propagated from a single edge (no gt)
            elif isinstance(edge, (TwoHopsNoGT, TwoHopsResidualNoGT)):
                validMessages = getTwoHopsMessages(edge, allMessages)
            else:
                assert False, f"Unknown type: {type(edge)}"
            # The algorithm for AAAI 2021 paper is computed as such only 1 message is propagated at most due to how
            #  the graph is constructed
            assert len(validMessages) <= len(self.cfg["inputNodes"]), (A, B, validMessages)
            for message in validMessages:
                y = edge.forward(message.output)
                newMessagePath = [*message.path, edge]
                newMessage = Message(newMessagePath, message.output, y)
                B.addMessage(newMessage)

    def forward(self, x):
        # NGC (V1) does only 2 graph iterations: InputNode->XXX (SL) and (InputNode->)XXX->YYY (TH).
        return super().forward(x, numIterations=2)

    def __str__(self):
        Str = super().__str__()
        Str += f"\n  -Input Nodes: {self.cfg['inputNodes']}"
        Str += "\n  -Single links:"
        for edge in filter(lambda edge: isinstance(edge, SingleLink), self.edges):
            Str += f"\n    -{edge}"

        Str += f"\n  -Two Hops (no GT):"
        for edge in filter(lambda edge: isinstance(edge, TwoHopsNoGT), self.edges):
            Str += f"\n    -{edge}"

        return Str


from nwgraph import Node, CriterionType
from typing import Dict

class NullNode(Node):
    def __init__(self, **kwargs):
        super().__init__(name="Null", hyperParameters={})

    def getNodeCriterion(self) -> CriterionType:
        return None
    
    def getNodeMetrics(self) -> Dict:
        return {}



import numpy as np
from media_processing_lib.image import image_resize_batch
from .deterministic_augmentation import DeterministicAugmentation
from ...logger import logger

class RandomZoom(DeterministicAugmentation):
    def __init__(self, percent_usage: float, max_percent_cut: float, seed: int=None, num_precomputted: int=100):
        super().__init__(seed, num_precomputted)
        self.max_percent_cut = max_percent_cut
        self.percent_usage = percent_usage

    def __call__(self, data):
        super().__call__()
        if np.random.random() > self.percent_usage / 100:
            logger.debug2("Skipping RandomZoom")
            return data

        # Generate the percentage tuples of top left and bottom right
        top_left_percent = np.random.random(size=(2, )) * self.max_percent_cut / 100
        bottom_right_percent = 1 - (np.random.random(size=(2, )) * self.max_percent_cut / 100)

        # Convert the data into a single concatenated array and remember where the initial indexes where
        split_ix = [y.shape[-1] for y in data[0: -1]]
        # MB x H x W x sum(ND)
        data = np.concatenate(data, axis=-1)
        h, w = data.shape[1 : 3]

        # Converting percentagers to pixel positions
        top_left = (top_left_percent * [h, w]).astype(int)
        bottom_right = (bottom_right_percent * [h, w]).astype(int)
        # Getting the zoomed data via slicing at the desired positions and resize back to original shape
        zoomed_data = data[:, top_left[0] : bottom_right[0], top_left[1] : bottom_right[1]]
        resized_data = image_resize_batch(zoomed_data, height=h, width=w, interpolation="nearest", only_uint8=False)

        # Split data back
        split_back = np.split(resized_data, split_ix, axis=-1)
        return split_back



import numpy as np
from nwgraph import Node

def defaultTransform(x: np.ndarray, node: Node) -> np.ndarray:
    x = np.float32(x)
    if len(x.shape) == 2:
        x = np.expand_dims(x, axis=-1)
    return x

def depthTransform(x: np.ndarray, node: Node) -> np.ndarray:
    if x.shape[-1] != 1:
        x = np.expand_dims(x, axis=-1)
    return np.float32(x)

def wireframeTransform(x: np.ndarray, node: Node) -> np.ndarray:
    x = x[..., 0 : 1]
    x = np.float32(x > 0)
    return x

def semanticSegmentationTransform(x: np.ndarray, node: Node) -> np.ndarray:
    semanticClasses = node.semanticClasses
    semanticClasses = semanticClasses if isinstance(semanticClasses, list) else list(range(semanticClasses))
    assert x.dtype in (np.uint8, np.uint16, np.uint32, np.int), x.dtype
    assert x.max() <= len(semanticClasses) - 1, f"{x.max()} vs {len(semanticClasses)}"
    if x.shape[-1] == 1:
        x = x[..., 0]
    # toCategorical
    y = np.eye(len(semanticClasses))[x]
    y = np.float32(y)
    return y



from typing import Callable
from functools import partial
import numpy as np
from matplotlib.cm import plasma
from media_processing_lib.image import to_image
from nwgraph import Node
from ..nodes import *

# Semantic data may come as argmaxed or one-hot, depending on how we store them (raw predictions
#  or compressed results).
def semanticToImage(x: np.ndarray, node: Semantic) -> np.ndarray:
	assert len(node.semanticClasses) == len(node.semanticColors)
	if len(x.shape) == 3 and x.shape[-1] != 1:
		assert len(node.semanticClasses) == x.shape[-1], f"{len(node.semanticClasses)} vs {x.shape}"
		x = np.argmax(x, axis=-1)
	nClasses = len(node.semanticClasses)
	assert x.max() < nClasses
	newImage = np.zeros((*x.shape, 3), dtype=np.uint8)
	for i in range(nClasses):
		newImage[x == i] = node.semanticColors[i]
	newImage = to_image(newImage)
	return newImage

def depthToImage(x: np.ndarray, node: Depth) -> np.ndarray:
	a = np.clip(x, 0, 1).squeeze()
	b = plasma(a)[..., 0 : 3]
	c = to_image(b)
	return c

def normalToImage(x: np.ndarray, node: Normal) -> np.ndarray:
	# TODO: if normals are not just unit vectors, but sin/cos or w/e else is predicted.
	return default(x, x)

def default(x: np.ndarray, node: Node) -> np.ndarray:
	x = np.clip(x, 0, 1)
	x = to_image(x)
	return x

def nullToBlackImage(x: np.ndarray, node: NullNode) -> np.ndarray:
	assert len(x.shape) == 3
	return np.zeros((x.shape[0], x.shape[1], 3), dtype=np.uint8)

def buildNodePlotFn(node: Node) -> Callable:
	if isinstance(node, RGB):
		fn = default
	elif isinstance(node, Depth):
		fn = depthToImage
	elif isinstance(node, Semantic):
		fn = semanticToImage
	elif isinstance(node, Normal):
		fn = normalToImage
	elif isinstance(node, NullNode):
		fn = nullToBlackImage
	else:
		assert False, f"Unknown node: '{node}' (type: {type(node)})"
	fn = partial(fn, node=node)
	return fn



from abc import abstractmethod, ABC
from typing import Dict
from nwgraph import Edge
from torch.utils.data import DataLoader

from ..readers import NGCNpzReader, EdgeReader
from ..models import NGC
from ..readers.augmentation import get_augmentation
from ..models.edges import SingleLink, TwoHopsNoGT, SingleLinkResidual, TwoHopsResidualNoGT

def get_loader(reader: NGCNpzReader, edge: Edge, train_cfg: Dict, randomize: bool, \
               debug: bool, augmentation: bool) -> DataLoader:
    """Gets a DataLoader from an NGCNpzReader, an edge and a train cfg"""
    general_augmentation = []
    node_specific_augmentation = None
    if augmentation and "augmentation" in train_cfg and "general" in train_cfg["augmentation"]:
        for name, args in train_cfg["augmentation"]["general"].items():
            general_augmentation.append(get_augmentation(name, **args))

    edge_reader = EdgeReader(reader.path, reader.nodes, edge, edge.getInKeys(), \
        general_augmentation, node_specific_augmentation)
    edge_loader = edge_reader.toDataLoader(batchSize=train_cfg["batchSize"], \
        randomize=randomize, debug=debug, seed=train_cfg["seed"])
    return edge_loader

class EdgeTrainer(ABC):
    def __init__(self, model: NGC):
        self.model = model

    @staticmethod
    def getEdgeDirName(edge: Edge) -> str:
        if isinstance(edge, SingleLink):
            suffix = f"{edge.inputNode}_{edge.outputNode}"
        elif isinstance(edge, TwoHopsNoGT):
            suffix = f"{edge.singleLinkNode}_{edge.inputNode}_{edge.outputNode}"
        elif isinstance(edge, SingleLinkResidual):
            suffix = f"{edge.residualNode}+{edge.inputNode}_{edge.outputNode}"
        elif isinstance(edge, TwoHopsResidualNoGT):
            suffix = f"{edge.residualNode}+{edge.singleLinkNode}_{edge.inputNode}_{edge.outputNode}"
        else:
            assert False, f"Unknown edge {edge} of type {type(edge)}."
        name = f"{edge.prefix}_{suffix}"
        return name

    @abstractmethod
    def run(self, trainCfg: Dict, trainReader: NGCNpzReader, validationReader: NGCNpzReader, debug: bool):
        pass
    
    def __call__(self, *args, **kwargs):
        self.run(*args, **kwargs)



import numpy as np
import torch as tr
from pathlib import Path
from typing import List
from ..readers import NGCNpzReader
from ..logger import logger, drange
from ..models.ngc import NGC
from ..nodes import Semantic

# @brief Assuming a pretrained model, we'll copy/link the labels of the training directory and then, for all
#  semisupervised directories, pass the data through the NGC model and save the results.
def exportPseudoLabels(model: NGC, semisupervisedDirs: List[Path], nextDataDir: Path, debug: bool):
	logger.debug(f"Exporting the trainDir and {len(semisupervisedDirs)} semisupervised dirs")
	logger.debug(f"Next data directory: '{nextDataDir}'")

	def exportOneDir(model: NGC, nextDataDir: Path, ssDir: Path, dirNumber: int, debug: bool):
		nodeNames = {node.name:node for node in model.nodes}
		inputNodes = [nodeNames[x] for x in model.cfg["inputNodes"]]
		reader = NGCNpzReader(path=ssDir, nodes=inputNodes, gtNodes=inputNodes)
		loader = reader.toDataLoader(randomize=False, debug=debug, numWorkers=0).__iter__()
		for j in drange(len(loader), desc="Pseudolabels"):
			x = next(loader)["data"]
			y = model.npForward(x)
			for nodeName, node in nodeNames.items():
				outDir = nextDataDir / nodeName
				assert outDir.exists()
				if nodeName in model.cfg["inputNodes"]:
					continue
				else:
					messages = y[node]
					assert len(messages) == 1, messages
					message = list(messages)[0]
					assert message.path[-1] == "Vote"
					nodeOutput = message.output[0]
					if isinstance(node, Semantic):
						nodeOutput = np.argmax(nodeOutput, axis=-1).astype(np.uint8)
				outFilePath = outDir / f"ss{dirNumber}_{j}.npz"
				if nodeOutput.dtype == np.float32:
					nodeOutput = nodeOutput.astype(np.float16)
				np.savez_compressed(outFilePath, nodeOutput)

	# Export pseudolabels for all semisupervised dirs (while copying the input nodes)
	logger.info("Exporting pseudolabels...")
	model.eval()
	for i in drange(len(semisupervisedDirs), desc="Pseudolabel dir"):
		ssDir = semisupervisedDirs[i]
		exportOneDir(model, nextDataDir, ssDir, dirNumber=i, debug=debug)

def export_new_dataset_every_iteration(ngcTrainer, iteration: int, debug: bool):
	"""The standard strategy, where for every new iteration, a new dataset is used to export pseudolabels.
	For all input nodes, there must be a reference gt label in semisupervised/iteration_x/node_name.
	"""
	from .edge_trainer import EdgeTrainer
	iterationModelsDir = ngcTrainer.ngcDir.getAllDataDirs(iteration)["models"][iteration]
	ngcTrainer.model.train()
	ngcTrainer.model.loadAllEdges(iterationModelsDir, EdgeTrainer.getEdgeDirName)
	ngcTrainer.model.eval()
	dataDir = ngcTrainer.ngcDir.getAllDataDirs(iteration + 1)["data"][iteration + 1]
	# Link all train labels
	ngcTrainer.dirLinker.linkTrainDir(iteration + 1)
	# Link all input nodes
	ngcTrainer.dirLinker.linkInputNodes(ngcTrainer.model.cfg["inputNodes"], iteration + 1)
	# Then export pseudolabels on top of train data
	exportPseudoLabels(ngcTrainer.model, ngcTrainer.dirLinker.semisupervisedDirs, dataDir, debug)

def export_train_set_only(ngcTrainer, iteration: int, debug: bool):
	# Just Link all train labels
	ngcTrainer.dirLinker.linkTrainDir(iteration + 1)

def export_pseudolabels_train_set_only(ngcTrainer, iteration: int, debug: bool):
	"""Method used by c-shift, where same dataset (train set) is iterated here. Input nodes are copied."""
	assert False, "TODO"
	ngcTrainer.dirLinker.linkTrainDir(iteration + 1)

def iterationPrologue(ngcTrainer, iteration: int, debug: bool):
	tr.use_deterministic_algorithms(False)
	method = ngcTrainer.method
	logger.info(f"Iteration {iteration - 1} prologue. Method: '{method}'")
	if method == "export_new_dataset_every_iteration":
		export_new_dataset_every_iteration(ngcTrainer, iteration, debug)
	elif method == "export_pseudolabels_train_set_only":
		export_pseudolabels_train_set_only(ngcTrainer, iteration, debug)
	elif method == "export_train_set_only":
		export_train_set_only(ngcTrainer, iteration, debug)
	else:
		assert False, f"Unknown method: '{method}'"



import os
from pathlib import Path
from natsort import natsorted
from typing import List
from .logger import logger

class DirLinker:
	def __init__(self, ngcDir, trainDir: Path, validationDir: Path, semisupervisedDirs: List[Path]):
		self.ngcDir = ngcDir
		self.trainDir = trainDir
		self.validationDir = validationDir
		self.semisupervisedDirs = semisupervisedDirs

	# @brief Creates a symlink between a training directory (given supervised list)
	#  and the data directory in the ngcdir
	def linkTrainDir(self, iteration: int):
		"""Creates a symlink between a training directory (given supervised data) and the data directory in the ngcdir
		"""
		nodeNames = self.ngcDir.getNodes()["names"]
		dataDir = self.ngcDir.path / f"iter{iteration}/data"
		logger.debug(f"Linking training dir {self.trainDir} -> {dataDir}")

		for nodeName in nodeNames:
			outDir = Path(dataDir / nodeName)
			if outDir.exists():
				logger.debug(f"Node out dir '{outDir}' exists. Skipping.")
				continue
			outDir.mkdir(exist_ok=False, parents=True)
			inDir = self.trainDir / nodeName
			inFiles = natsorted([str(x) for x in inDir.glob("*.npz")])
			outFiles = [outDir / f"train_{i}.npz" for i in range(len(inFiles))]
			for inFile, outFile in zip(inFiles, outFiles):
				os.symlink(inFile, outFile)
		logger.debug(f"Finished linking training dir.")

	def linkInputNodes(self, inputNodes: List[str], iteration: int):
		"""Links all inpiut nodes for this iteration based on the input nodes list."""
		dataDir = self.ngcDir.path / f"iter{iteration}/data"
		logger.debug(f"Linking semisupervised dirs ({len(self.semisupervisedDirs)}) -> {dataDir} for all "
			f"input nodes: {inputNodes}")
		dataDir = self.ngcDir.path / f"iter{iteration}/data"
		for j, ssDir in enumerate(self.semisupervisedDirs):
			for nodeName in inputNodes:
				assert (dataDir / nodeName).exists(), dataDir / nodeName
				assert (ssDir / nodeName).exists(), ssDir / nodeName
				inFiles = natsorted([str(x) for x in (ssDir / nodeName).glob("*.npz")])
				outFiles = [dataDir / nodeName / f"ss{j}_{i}.npz" for i in range(len(inFiles))]
				for inFile, outFile in zip(inFiles, outFiles):
					if not outFile.exists():
						os.symlink(inFile, outFile)
		logger.debug(f"Finished linking input nodes for semisupervised dirs.")



from copy import deepcopy
from typing import Dict, List
import torch as tr
from overrides import overrides
from torch import nn
from nwmodule import NWModule
from nwutils.nwmodule import trModuleWrapper
from nwutils.torch import trGetData
from nwgraph import Node
from ..edges import NGCEdge
from ...logger import logger

class SingleLinkResidual(NGCEdge):
    def __init__(self, residualNode: Node, inputNode: Node, outputNode: Node, hyperParameters: Dict={}):
        hyperParameters["residualNode"] = residualNode
        name = f"Single Link Residual ({inputNode} -> {outputNode}) (Residual Node: {residualNode})"
        prefix = "SLR"
        super().__init__(prefix, inputNode, outputNode, name, hyperParameters)
        self.residualNode = residualNode

    @overrides
    def getModel(self) -> NWModule:
        if hasattr(self, "model"):
            logger.info("Model already instantiated, returning early.")
            return self.model
        A, B = self.inputNode, self.outputNode
        # Create a new intermediate node where we append the SL's (RGB) numbers of dimensions to instantaite the model
        #  properly.
        _A = deepcopy(A)
        _A.numDims += self.hyperParameters["residualNode"].numDims
        encoder = _A.getEncoder(B)
        decoder = B.getDecoder(A)
        model = trModuleWrapper(nn.Sequential(encoder, decoder))
        self.addMetrics(B.getNodeMetrics())
        return model

    @overrides
    def networkAlgorithm(self, trInputs, trLabels, isTraining, isOptimizing):
        """Called at train time, graph is not passed, but we have access bot to SL input as well as residual. """
        assert not self.residualNode is None
        assert not self.getCriterion() is None, "Set criterion before training or testing"
        # Pass through the single link of this two hops edge
        residual = trInputs[self.residualNode]
        edgeInput = trInputs[self.inputNode]
        concatenated = tr.cat([edgeInput, residual], dim=-1)
        trResults = self.model.forward(concatenated)
        trLoss = 0
        if isTraining:
            trLoss = self.getCriterion()(trResults, trLabels)
            self.updateOptimizer(trLoss, isTraining, isOptimizing)
        return trResults, trLoss

    def forward(self, x: tr.Tensor):
        """Called only at inference time, so graph is passed. """
        assert len(self.residualNode.messages) == 1
        residualInput = trGetData(tuple(self.residualNode.messages)[0].input)
        concatenated = tr.cat([x, residualInput], dim=-1)
        y = super().forward(concatenated)
        return y

    @overrides
    def getInKeys(self) -> List[str]:
        return [self.inputNode.name, self.residualNode.name]



from typing import List, Dict
from overrides import overrides
from nwgraph import Node
from .ngc_edge import NGCEdge

class SingleLink(NGCEdge):
    """Simple Single Link edge. Defined by a regular forward pass."""
    def __init__(self, inputNode: Node, outputNode: Node, hyperParameters: Dict = {}):
        name = f"Single Link ({inputNode} -> {outputNode})"
        prefix = "SL"
        super().__init__(prefix, inputNode, outputNode, name, hyperParameters)

    @overrides
    def networkAlgorithm(self, trInputs, trLabels, isTraining, isOptimizing):
        """Called at train time, graph is not passed, but we have access to proper GT (SL input). """
        assert not self.getCriterion() is None, "Set criterion before training or testing"
        # Pass through the single link of this two hops edge
        edgeInput = trInputs[self.inputNode]
        trResults = self.model.forward(edgeInput)
        trLoss = 0
        if isTraining:
            trLoss = self.getCriterion()(trResults, trLabels)
            self.updateOptimizer(trLoss, isTraining, isOptimizing)
        return trResults, trLoss

    @overrides
    def getInKeys(self) -> List[str]:
        return [self.inputNode.name]



from copy import deepcopy
from typing import Dict, List
import torch as tr
from pathlib import Path
from overrides import overrides
from torch import nn
from nwmodule import NWModule
from nwutils.nwmodule import trModuleWrapper
from nwutils.torch import trGetData
from nwgraph import Edge, Node
from .two_hops_nogt import TwoHopsSerializer
from .ngc_edge import NGCEdge
from ...logger import logger

class TwoHopsResidualNoGT(NGCEdge):
    def __init__(self, singleLinkEdge: Edge, residualNode: Node, \
            inputNode: Node, outputNode: Node, hyperParameters: Dict={}):
        singleLinkNode = singleLinkEdge.inputNode
        name = f"Two Hops NoGT ({inputNode} -> {outputNode}) (Single Link Node: {singleLinkNode}. " \
            f"Residual Node: {residualNode})"
        prefix = "THR_NOGT"
        hyperParameters["singleLinkNode"] = singleLinkNode
        hyperParameters["residualNode"] = residualNode
        super().__init__(prefix, inputNode, outputNode, name, hyperParameters)

        self.singleLinkNode = singleLinkNode
        self.residualNode = residualNode
        self.singleLinkEdge = deepcopy(singleLinkEdge).to(self.getDevice())
        self.serializer = TwoHopsSerializer(model=self)
        logger.debug(f"Setting single link of {self.name}: '{self.singleLinkEdge}'")

    @overrides
    def getModel(self) -> NWModule:
        if hasattr(self, "model"):
            logger.info("Model already instantiated, returning early.")
            return self.model
        A, B = self.inputNode, self.outputNode
        # Create a new intermediate node where we append the SL's (RGB) numbers of dimensions to instantaite the model
        #  properly.
        _A = deepcopy(A)
        _A.numDims += self.hyperParameters["residualNode"].numDims
        encoder = _A.getEncoder(B)
        decoder = B.getDecoder(A)
        model = trModuleWrapper(nn.Sequential(encoder, decoder))
        self.addMetrics(B.getNodeMetrics())
        return model

    @overrides
    def setup(self, **kwargs):
        weightsFile = f"{kwargs['singleLinkDir']}/model_best_Loss.pkl" 
        logger.debug(f"Loading single link of {self.name}: {self.singleLinkEdge} from '{weightsFile}'")
        self.singleLinkEdge.setTrainableWeights(True)
        self.singleLinkEdge.loadWeights(weightsFile)
        self.singleLinkEdge.eval()

    @overrides
    def networkAlgorithm(self, trInputs, trLabels, isTraining, isOptimizing):
        """Called at train time, graph is not passed, but we have access to proper GT (SL input). """
        assert not self.singleLinkEdge is None
        assert not self.getCriterion() is None, "Set criterion before training or testing"
        # Pass through the single link of this two hops edge
        residual = trInputs[self.residualNode.name]
        with tr.no_grad():
            slOutput, _ = self.singleLinkEdge.networkAlgorithm(trInputs, trLabels, False, False)
        concatenated = tr.cat([slOutput, residual], dim=-1)
        trResults = self.model.forward(concatenated)
        trLoss = 0
        if isTraining:
            trLoss = self.getCriterion()(trResults, trLabels)
            self.updateOptimizer(trLoss, isTraining, isOptimizing)
        return trResults, trLoss

    def forward(self, x: tr.Tensor):
        """Called only at inference time, so graph is passed. """
        assert len(self.residualNode.messages) == 1
        slInput = trGetData(tuple(self.residualNode.messages)[0].input)
        concatenated = tr.cat([x, slInput], dim=-1)
        y = super().forward(concatenated)
        return y

    @overrides
    def getInKeys(self) -> List[str]:
        return self.singleLinkEdge.getInKeys()




import sys
import yaml
import importlib
import torch as tr
from typing import List
from pathlib import Path
from nwmodule import NWModule
from nwgraph import Node
from ..logger import logger

device = tr.device("cuda") if tr.cuda.is_available() else tr.device("cpu")

class NetworkSelection:
    def __init__(self, vote:List):
        assert len(vote) == 5
        assert vote[0] == "networkSelection"
        assert Path(vote[2]).exists(), f"Vote: Path to model module '{vote[2]}' doesn't exist"
        assert Path(vote[3]).exists(), f"Vote: Path to model cfg '{vote[3]}' doesn't exist"
        assert Path(vote[4]).exists(), f"Vote: Path to model weights '{vote[4]}' doesn't exist"
        logger.debug(f"Loading model from module '{vote[2]}', model '{vote[1]}', cfg '{vote[3]}', weights '{vote[4]}'")
        self.vote = vote
        self.model = self.getModel(vote)

    def getModel(self, vote:List) -> NWModule:
        _, modelName, modelModulePath, modelCfg, modelWeights = vote
        spec = importlib.util.spec_from_file_location("module.name", modelModulePath)
        sys.path.append(str(Path(modelModulePath).parents[0]))
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        modelType = getattr(foo, modelName)
        model = modelType(yaml.safe_load(open(modelCfg, "r")))
        model.setTrainableWeights(True)
        model.loadWeights(modelWeights)
        model.eval()
        model.to(device)
        return model

    def __call__(self, node:Node, messages:tr.Tensor):
        return self.model(node, messages)



import numpy as np
from overrides import overrides
from typing import List, Union, Dict
from nwgraph import Edge, Node
from .ngc_npz_reader import NGCNpzReader
from .augmentation import NGCAugmentation, GeneralPrototype, NodeSpecificPrototype, get_augmentation

# Reader used to train a single edge at a time. For now this reader takes the original reader (with all possible gtKeys
#  for the graph) and only sends the data and labels required for this particular edge. We assume that edges only need
#  an input. More complicated edges might need additional readers.
class EdgeReader(NGCNpzReader):
	def __init__(self, path, nodes: List[Node], edge: Edge, inKeys: Union[Node, List[Node]], \
				 general_augmentation: List[GeneralPrototype]=None, \
                 node_specific_augmentation: Dict[str, List[NodeSpecificPrototype]]=None):
		if isinstance(inKeys, str):
			inKeys = [inKeys]
		# We only care about all input keys of the edge plus the output node of the edge (for gt)
		node_name_to_node = {node.name: node for node in nodes}
		gt_node = node_name_to_node[edge.outputNode.name]
		relevant_node_names = [*inKeys, edge.outputNode.name]
		relevant_nodes = [node_name_to_node[x] for x in relevant_node_names]
	
		self.augmentation = NGCAugmentation(relevant_nodes, general_augmentation, node_specific_augmentation)
		super().__init__(path=path, nodes=relevant_nodes, gtNodes=[gt_node])
		self.edge = edge
		self.inKeys = inKeys

	@overrides
	def mergeFn(self, x):
		"""EdgeReader will not behave like the graph reader. Data is given as a dict, while labels as a regular
		array. Edges will have to understand this format for training.
		"""
		res = super().mergeFn(x)
		all_nodes_data = res["data"]
		all_nodes_data = self.augmentation(all_nodes_data)
		data = {k: all_nodes_data[k] for k in self.inKeys}
		labels = all_nodes_data[self.edge.outputNode.name]
		return {"data": data, "labels": labels}

	def __str__(self) -> str:
		A, B = self.edge.getNodes()
		Type = str(type(self.edge)).split(".")[-1][0 : -2]
		Str = f"[EdgeReader] {A}->{B} (Edge Type: {Type} inKeys: {self.inKeys})"
		Str += super().__str__()
		return Str
